/*
    SDL - Simple DirectMedia Layer
    Copyright (C) 1997, 1998, 1999  Sam Lantinga

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Sam Lantinga
    slouken@devolution.com
*/

#ifdef SAVE_RCSID
static char rcsid =
 "@(#) $Id: SDL_x11modes.c,v 1.6 2000/01/25 00:14:42 hercules Exp $";
#endif

/* Utilities for getting and setting the X display mode */

#include "SDL_x11video.h"
#include "SDL_x11modes_c.h"


#ifdef XFREE86_VM
Bool XF86VidModeGetModeInfo(Display *dpy, int scr, XF86VidModeModeInfo *info)
{
    XF86VidModeModeLine *l = (XF86VidModeModeLine*)((char*)info + sizeof info->dotclock);
    return XF86VidModeGetModeLine(dpy, scr, &info->dotclock, l);
}
#endif /* XFREE86_VM */

static void save_mode(_THIS)
{
#ifdef XFREE86_VM
    memset(&saved_mode, 0, sizeof(saved_mode));
    XF86VidModeGetModeInfo(SDL_Display,SDL_Screen,&saved_mode);
    XF86VidModeGetViewPort(SDL_Display,SDL_Screen,&saved_view.x,&saved_view.y);
#endif
}

static void restore_mode(_THIS)
{
#ifdef XFREE86_VM
    XF86VidModeModeLine mode;
    int unused;

    if ( XF86VidModeGetModeLine(SDL_Display, SDL_Screen, &unused, &mode) ) {
        if ( (saved_mode.hdisplay != mode.hdisplay) ||
             (saved_mode.vdisplay != mode.vdisplay) ) {
            XF86VidModeSwitchToMode(SDL_Display, SDL_Screen, &saved_mode);
        }
    }
    if ( (saved_view.x != 0) || (saved_view.y != 0) ) {
        XF86VidModeSetViewPort(SDL_Display, SDL_Screen, saved_view.x, saved_view.y);
    }
#endif
}

static int cmpmodes(const void *va, const void *vb)
{
#ifdef XFREE86_VM
    XF86VidModeModeInfo *a = *(XF86VidModeModeInfo**)va;
    XF86VidModeModeInfo *b = *(XF86VidModeModeInfo**)vb;
    if(a->hdisplay > b->hdisplay)
        return -1;
    return b->vdisplay - a->vdisplay;
#endif
}

static void set_best_resolution(_THIS, int width, int height)
{
#ifdef XFREE86_VM
    if ( use_vidmode ) {
        XF86VidModeModeLine mode;
        XF86VidModeModeInfo **modes;
        int i;
        int nmodes;

        if ( XF86VidModeGetModeLine(SDL_Display, SDL_Screen, &i, &mode) &&
             XF86VidModeGetAllModeLines(SDL_Display,SDL_Screen,&nmodes,&modes)){
            qsort(modes, nmodes, sizeof *modes, cmpmodes);
#ifdef XFREE86_DEBUG
  printf("Available modes:\n");
  for ( i = 0; i < nmodes; ++i ) {
    printf("Mode %d: %dx%d\n", i, modes[i]->hdisplay, modes[i]->vdisplay);
  }
#endif
            for ( i = nmodes-1; i >= 0 ; --i ) {
                if ( (modes[i]->hdisplay >= width) &&
                     (modes[i]->vdisplay >= height) )
                    break;
            }
            if ( i >= 0 ) {
                if ( (modes[i]->hdisplay != mode.hdisplay) ||
                     (modes[i]->vdisplay != mode.vdisplay) ) {
                    XF86VidModeSwitchToMode(SDL_Display, SDL_Screen, modes[i]);
                }
            }
            XFree(modes);
        }
    }
#endif /* XFREE86_VM */
}

static void get_real_resolution(_THIS, int* w, int* h)
{
#ifdef XFREE86_VM
    if ( use_vidmode ) {
        XF86VidModeModeLine mode;
        int unused;

        if ( XF86VidModeGetModeLine(SDL_Display, SDL_Screen, &unused, &mode) ) {
            *w = mode.hdisplay;
            *h = mode.vdisplay;
            return;
        }
    }
#endif
    *w = DisplayWidth(SDL_Display, SDL_Screen);
    *h = DisplayHeight(SDL_Display, SDL_Screen);
}

/* Called after mapping a window - waits until the window is mapped */
void X11_WaitMapped(_THIS, Window win)
{
    XEvent event;
    do {
        XMaskEvent(SDL_Display, StructureNotifyMask, &event);
    } while ( (event.type != MapNotify) || (event.xmap.event != win) );
}

/* Called after unmapping a window - waits until the window is unmapped */
void X11_WaitUnmapped(_THIS, Window win)
{
    XEvent event;
    do {
        XMaskEvent(SDL_Display, StructureNotifyMask, &event);
    } while ( (event.type != UnmapNotify) || (event.xunmap.event != win) );
}

static void move_cursor_to(_THIS, int x, int y)
{
    XWarpPointer(SDL_Display, None, SDL_Root, 0, 0, 0, 0, x, y);
}

void X11_GetVideoModes(_THIS)
{
#ifdef XFREE86_VM
    int buggy_X11;
    int vm_event, vm_error;
    int major_version, minor_version;
    int nmodes;
    XF86VidModeModeInfo **modes;
#endif
    int i;

    use_vidmode = 0;
#ifdef XFREE86_VM
    /* Metro-X 4.3.0 and earlier has a broken implementation of
       XF86VidModeGetAllModeLines() - it hangs the client.
     */
    buggy_X11 = 0;
    if ( strcmp(ServerVendor(SDL_Display), "Metro Link Incorporated") == 0 ) {
        FILE *metro_fp;

        metro_fp = fopen("/usr/X11R6/lib/X11/Metro/.version", "r");
        if ( metro_fp != NULL ) {
            int major, minor, patch, version;
            major = 0; minor = 0; patch = 0;
            fscanf(metro_fp, "%d.%d.%d", &major, &minor, &patch);
            version = major*100+minor*10+patch;
            if ( version < 431 ) {
                buggy_X11 = 1;
            }
            fclose(metro_fp);
        }
    }
#if defined(__alpha__) || defined(__powerpc__)
    /* The alpha and PPC XFree86 servers are also buggy */
    buggy_X11 = 1;
#endif
    /* Enumerate the available fullscreen modes */
    if ( local_X11 && ! buggy_X11 &&
         XF86VidModeQueryExtension(SDL_Display, &vm_event, &vm_error) &&
         XF86VidModeQueryVersion(SDL_Display, &major_version, &minor_version) &&
         XF86VidModeGetAllModeLines(SDL_Display, SDL_Screen,&nmodes,&modes) ) {

        qsort(modes, nmodes, sizeof *modes, cmpmodes);
        SDL_modelist = (SDL_Rect **)malloc((nmodes+1)*sizeof(SDL_Rect *));
        if ( SDL_modelist ) {
            for ( i=0; i<nmodes; ++i ) {
                SDL_modelist[i] = (SDL_Rect *)malloc(sizeof(SDL_Rect));
                if ( SDL_modelist[i] == NULL ) {
                    break;
                }
                SDL_modelist[i]->x = 0;
                SDL_modelist[i]->y = 0;
                SDL_modelist[i]->w = modes[i]->hdisplay;
                SDL_modelist[i]->h = modes[i]->vdisplay;
            }
            SDL_modelist[i] = NULL;
        }
        XFree(modes);

        use_vidmode = 1;
        save_mode(this);
    }
#endif /* XFREE86_VM */

    if ( SDL_modelist == NULL ) {
        SDL_modelist = (SDL_Rect **)malloc((1+1)*sizeof(SDL_Rect *));
        i = 0;
        if ( SDL_modelist ) {
            SDL_modelist[i] = (SDL_Rect *)malloc(sizeof(SDL_Rect));
            if ( SDL_modelist[i] ) {
                SDL_modelist[i]->x = 0;
                SDL_modelist[i]->y = 0;
                SDL_modelist[i]->w = DisplayWidth(SDL_Display, SDL_Screen);
                SDL_modelist[i]->h = DisplayHeight(SDL_Display, SDL_Screen);
                ++i;
            }
            SDL_modelist[i] = NULL;
        }
    }

#ifdef DEBUG_XF86
    if ( use_vidmode ) {
        fprintf(stderr, "XFree86 VidMode is enabled\n");
    }
    if ( SDL_modelist ) {
        fprintf(stderr, "X11 video mode list:\n");
        for ( i=0; SDL_modelist[i]; ++i ) {
            fprintf(stderr, "\t%dx%d\n",
                SDL_modelist[i]->w, SDL_modelist[i]->h);
        }
    }
#endif /* DEBUG_XF86 */
}

int X11_SupportedVisual(_THIS, SDL_PixelFormat *format)
{
    return(format->BitsPerPixel == this->screen->format->BitsPerPixel);
}

SDL_Rect **X11_ListModes(_THIS, SDL_PixelFormat *format, Uint32 flags)
{
    if ( X11_SupportedVisual(this, format) ) {
        if ( flags & SDL_FULLSCREEN ) {
            return(SDL_modelist);
        } else {
            return((SDL_Rect **)-1);
        }
    } else {
        return((SDL_Rect **)0);
    }
}

void X11_FreeVideoModes(_THIS)
{
    int i;

    if ( SDL_modelist ) {
        for ( i=0; SDL_modelist[i]; ++i ) {
            free(SDL_modelist[i]);
        }
        free(SDL_modelist);
        SDL_modelist = NULL;
    }
}

int X11_ResizeFullScreen(_THIS)
{
    int x, y;
    int real_w, real_h;

    if ( currently_fullscreen ) {
        /* Switch resolution and cover it with the FSwindow */
        move_cursor_to(this, 0, 0);
        set_best_resolution(this, current_w, current_h);
        move_cursor_to(this, 0, 0);
        get_real_resolution(this, &real_w, &real_h);
        XResizeWindow(SDL_Display, FSwindow, real_w, real_h);
        move_cursor_to(this, real_w/2, real_h/2);

        /* Center and reparent the drawing window */
        x = (real_w - current_w)/2;
        y = (real_h - current_h)/2;
        XReparentWindow(SDL_Display, SDL_Window, FSwindow, x, y);
        /* FIXME: move the mouse to the old relative location */
        XSync(SDL_Display, True);   /* Flush spurious mode change events */
    }
    return(1);
}

void X11_QueueEnterFullScreen(_THIS)
{
    switch_waiting = 0x01 | SDL_FULLSCREEN;
    switch_time = SDL_GetTicks() + 1500;
    XSetInputFocus(SDL_Display, WMwindow, RevertToNone, CurrentTime);
}

int X11_EnterFullScreen(_THIS)
{
    int okay;
    Window tmpwin, *windows;
    int i, nwindows;

    okay = 1;
    if ( ! currently_fullscreen ) {
        int real_w, real_h;

        /* Map the fullscreen window to blank the screen */
        get_real_resolution(this, &real_w, &real_h);
        XResizeWindow(SDL_Display, FSwindow, real_w, real_h);
        XMapRaised(SDL_Display, FSwindow);
        X11_WaitMapped(this, FSwindow);

        /* Make sure we got to the top of the window stack */
        if ( XQueryTree(SDL_Display, SDL_Root, &tmpwin, &tmpwin,
                                &windows, &nwindows) && windows ) {
            /* If not, try to put us there - if fail... oh well */
            if ( windows[nwindows-1] != FSwindow ) {
                tmpwin = windows[nwindows-1];
                for ( i=0; i<nwindows; ++i ) {
                    if ( windows[i] == FSwindow ) {
                        memcpy(&windows[i], &windows[i+1],
                               (nwindows-i-1)*sizeof(windows[i]));
                        break;
                    }
                }
                windows[nwindows-1] = FSwindow;
                XRestackWindows(SDL_Display, windows, nwindows);
                XSync(SDL_Display, False);
            }
            XFree(windows);
        }

        /* Grab the mouse on the fullscreen window
           The event handling will know when we become active, and then
           enter fullscreen mode if we can't grab the mouse this time.
         */
#ifdef GRAB_FULLSCREEN
        if ( (XGrabPointer(SDL_Display, FSwindow, True, 0,
                          GrabModeAsync, GrabModeAsync,
                          FSwindow, None, CurrentTime) != GrabSuccess) ||
             (XGrabKeyboard(SDL_Display, WMwindow, True,
                          GrabModeAsync, GrabModeAsync, CurrentTime) != 0) ) {
#else
        if ( XGrabPointer(SDL_Display, FSwindow, True, 0,
                          GrabModeAsync, GrabModeAsync,
                          FSwindow, None, CurrentTime) != GrabSuccess ) {
#endif
            /* We lost the grab, so try again later */
            XUnmapWindow(SDL_Display, FSwindow);
            X11_WaitUnmapped(this, FSwindow);
            X11_QueueEnterFullScreen(this);
            return(0);
        }

#ifdef XFREE86_VM
        /* Save the current video mode */
        if ( use_vidmode ) {
            XF86VidModeLockModeSwitch(SDL_Display, SDL_Screen, True);
        }
#endif
        currently_fullscreen = 1;

        /* Set the new resolution */
        okay = X11_ResizeFullScreen(this);
        if ( ! okay ) {
            X11_LeaveFullScreen(this);
        }
    }
    X11_GrabInputNoLock(this, this->input_grab);
    return(okay);
}

int X11_LeaveFullScreen(_THIS)
{
    if ( currently_fullscreen ) {
        XReparentWindow(SDL_Display, SDL_Window, WMwindow, 0, 0);
#ifdef XFREE86_VM
        if ( use_vidmode ) {
            restore_mode(this);
            XF86VidModeLockModeSwitch(SDL_Display, SDL_Screen, False);
        }
#endif
        XUnmapWindow(SDL_Display, FSwindow);
        X11_WaitUnmapped(this, FSwindow);
#ifdef GRAB_FULLSCREEN
        XUngrabKeyboard(SDL_Display, CurrentTime);
#endif
        XSync(SDL_Display, True);   /* Flush spurious mode change events */
        currently_fullscreen = 0;
    }
    X11_GrabInputNoLock(this, this->input_grab);
}
