#include "../libmpeg2.h"
#include "../mpeg2private.h"
#include "../mpeg2protos.h"
#include "mpeg2audio.h"
#include "tables.h"

#include <math.h>
#include <stdlib.h>

mpeg2audio_t* mpeg2audio_allocate_struct(mpeg2_t *file, mpeg2_atrack_t *track)
{
	mpeg2audio_t *audio = malloc(sizeof(mpeg2audio_t));
	audio->file = file;
	audio->track = track;
	audio->astream = mpeg2bits_new_stream(file, track->demuxer);
	audio->outscale = 1;
	audio->bsbufold = 0;
	audio->bsbuf = audio->bsspace[1];
	audio->bsnum = 0;
	audio->lsf = 0;
	audio->mpeg25 = 0;
	audio->lay = 0;
	audio->init = 1;
	audio->sampling_frequency_code = 0;
	audio->framenum = 0;
	audio->prev_framesize = 0;
	audio->single = 0;
	audio->pcm_sample = 0;
	audio->pcm_point = 0;
	audio->pcm_size = 0;
	audio->pcm_allocated = 0;
	audio->bo = 1;
	return audio;
}


int mpeg2audio_delete_struct(mpeg2audio_t *audio)
{
	mpeg2bits_delete_stream(audio->astream);
	if(audio->pcm_sample) free(audio->pcm_sample);
	free(audio);
	return 0;
}

int mpeg2audio_replace_buffer(mpeg2audio_t *audio, long new_allocation)
{
	long i;

	if(!audio->pcm_sample)
	{
		audio->pcm_sample = malloc(sizeof(float) * new_allocation * audio->stereo);
		audio->pcm_allocated = new_allocation;
	}
	else
	{
		float *new_samples = malloc(sizeof(float) * new_allocation * audio->stereo);
		for(i = 0; i < audio->pcm_size * audio->stereo; i++)
		{
			new_samples[i] = audio->pcm_sample[i];
		}
		free(audio->pcm_sample);
		audio->pcm_sample = new_samples;
		audio->pcm_allocated = new_allocation;
	}
	return 0;
}

int mpeg2audio_read_frame(mpeg2audio_t *audio)
{
	int result = 0;
	result = mpeg2audio_read_header(audio);

	if(!result)
	{
		switch(audio->lay)
		{
			case 1:
				break;

			case 2:
				result = mpeg2audio_dolayer2(audio);
				break;

			case 3:
				result = mpeg2audio_dolayer3(audio);
				break;

			default:
				result = 1;
				break;
		}
		if(!result)
		{
// Byte align the stream
			mpeg2bits_byte_align(audio->astream);
		}
	}
	return result;
}

// Get the length but also initialize the frame sizes.

int mpeg2audio_get_length(mpeg2audio_t *audio, mpeg2_atrack_t *track)
{
	long result = 0;
	long framesize1 = 0, total1 = 0;
	long framesize2 = 0, total2 = 0;
	long total_framesize = 0, total_frames = 0;
	long byte_limit = 65535;  // Total bytes to gather information from
	long total_bytes = 0;
	long major_framesize;     // Bigger framesize + header
	long minor_framesize;     // Smaller framesize + header
	long major_total;
	long minor_total;
	mpeg2_t *file = audio->file;

// Get the frame sizes
	mpeg2bits_seek_start(audio->astream);
	audio->pcm_point = 0;
	result = mpeg2audio_read_frame(audio);
	audio->samples_per_frame = audio->pcm_point / audio->stereo;
	framesize1 = audio->framesize;
	total_bytes += audio->framesize;
	total1 = 1;

	while(!result && total_bytes < byte_limit)
	{
		audio->pcm_point = 0;
		result = mpeg2audio_read_frame(audio);
		total_bytes += audio->framesize;
		if(audio->framesize != framesize1)
		{
			framesize2 = audio->framesize;
			total2 = 1;
			break;
		}
		else
		{
			total1++;
		}
	}

	while(!result && total_bytes < byte_limit)
	{
		audio->pcm_point = 0;
		result = mpeg2audio_read_frame(audio);
		total_bytes += audio->framesize;
		if(audio->framesize != framesize2)
		{
			break;
		}
		else
		{
			total2++;
		}
	}

	audio->pcm_point = 0;
	result = mpeg2audio_read_frame(audio);
	if(audio->framesize != framesize1 && audio->framesize != framesize2)
	{
// Variable bit rate.  Get the average frame size.
		while(!result && total_bytes < byte_limit)
		{
			audio->pcm_point = 0;
			result = mpeg2audio_read_frame(audio);
			total_bytes += audio->framesize;
			if(!result)
			{
				total_framesize += audio->framesize;
				total_frames++;
			}
		}
		audio->avg_framesize = 4 + (float)(total_framesize + framesize1 + framesize2) / (total_frames + total1 + total2);
	}
	else
	{
		major_framesize = framesize2 > framesize1 ? framesize2 : framesize1;
		major_total = framesize2 > framesize1 ? total2 : total1;
		minor_framesize = framesize2 > framesize1 ? framesize1 : framesize2;
		minor_total = framesize2 > framesize1 ? total1 : total2;
// Add the headers to the framesizes
		audio->avg_framesize = 4 + (float)(major_framesize * major_total + minor_framesize * minor_total) / (major_total + minor_total);
	}

	audio->pcm_point = 0;
	mpeg2bits_seek_start(audio->astream);
	mpeg2audio_reset_synths(audio);

// Estimate the total samples
	if(file->is_audio_stream)
	{
// From the raw file
		result = (long)((float)file->total_bytes / audio->avg_framesize * audio->samples_per_frame);
	}
	else
	{
// Gross approximation from a multiplexed file.
		result = (long)((float)mpeg2_video_frames(file) / mpeg2_frame_rate(file) * track->sample_rate);
// We would scan the multiplexed packets here for the right timecode if only
// they had meaningful timecode.
	}
	return result;
}

int mpeg2audio_seek(mpeg2audio_t *audio, long position)
{
	int result = 0;
	mpeg2_t *file = audio->file;
	mpeg2_atrack_t *track = audio->track;
	long frame_number;
	long byte_position;

// Can't slide buffer.  Seek instead.
	if(!file->is_audio_stream)
	{
// Seek in a multiplexed stream.  Gross approximation.
		byte_position = (long)((float)position / track->total_samples * file->total_bytes);
// We would scan the multiplexed packets here for the right timecode if only
// they had meaningful timecode.
	   	audio->pcm_position = position;
	}
	else
	{
// Seek in an elemental stream.  This algorithm achieves sample accuracy on fixed bitrates.
// Forget about variable bitrates or program streams.
		frame_number = position / audio->samples_per_frame;
		byte_position = (long)(audio->avg_framesize * frame_number);
	   	audio->pcm_position = frame_number * audio->samples_per_frame;
	}

	if(byte_position < audio->avg_framesize * 2)
	{
		result |= mpeg2bits_seek_start(audio->astream);
		audio->pcm_position = 0;
	}
	else
	{
		result |= mpeg2bits_seek_stream(audio->astream, byte_position);
	}

// Arm the backstep buffer for layer 3 if not at the beginning already.
	if(byte_position >= audio->avg_framesize * 2 && audio->lay == 3 && !result)
	{
		result |= mpeg2audio_prev_header(audio);
		result |= mpeg2audio_read_layer3_frame(audio);
	}

// Reset the tables.
	mpeg2audio_reset_synths(audio);
	audio->pcm_size = 0;
	audio->pcm_point = 0;
	return result;
}

// ================================================================
//                                    ENTRY POINTS
// ================================================================




mpeg2audio_t* mpeg2audio_new(mpeg2_t *file, mpeg2_atrack_t *track)
{
	mpeg2audio_t *audio = mpeg2audio_allocate_struct(file, track);
	int result = 0;

// Init tables
	mpeg2audio_new_decode_tables(audio);
	mpeg2audio_init_layer2(audio); /* inits also shared tables with layer1 */
	mpeg2audio_init_layer3(audio);
	result = mpeg2audio_read_header(audio);
// Layer 1 not supported
	if(audio->lay == 1) 
	{
		fprintf(stderr, "mpeg2audio_new: layer 1 not supported\n");
		result = 1;
	}

// Set up the sample buffer
	mpeg2audio_replace_buffer(audio, MPEG2AUDIO_PADDING);
	if(!result)
	{
		track->channels = audio->stereo;
		track->sample_rate = mpeg2_freqs[audio->sampling_frequency_code];
		track->total_samples = mpeg2audio_get_length(audio, track);
	}
	else
	{
		mpeg2audio_delete_struct(audio);
		audio = 0;
	}

	return audio;
}

int mpeg2audio_delete(mpeg2audio_t *audio)
{
	mpeg2audio_delete_struct(audio);
	return 0;
}

int mpeg2audio_read_raw(mpeg2audio_t *audio, unsigned char *output, long *size, long max_size)
{
	int result = 0;
	*size = 0;
	result = mpeg2audio_read_header(audio);
	if(!result)
	{
		if(max_size < 4) return 1;
		*output++ = (audio->newhead & 0xff000000) >> 24;
		*output++ = (audio->newhead & 0xff0000) >> 16;
		*output++ = (audio->newhead & 0xff00) >> 8;
		*output++ = (audio->newhead & 0xff);
		*size += 4;
		
		if(max_size < 4 + audio->framesize) return 1;
		if(mpeg2bits_read_buffer(audio->astream, output, audio->framesize))
			return 1;
		
		*size += audio->framesize;
	}
	return result;
}

// Channel is 0 or 1
int mpeg2audio_decode_audio(mpeg2audio_t *audio, float *output_f, short *output_i, int channel, long start_position, long len)
{
	long allocation_needed = len + MPEG2AUDIO_PADDING;
	long i, j, result = 0;
	mpeg2_t *file = audio->file;
	mpeg2_atrack_t *atrack = audio->track;

// Create new buffer
	if(audio->pcm_allocated < allocation_needed)
	{
		mpeg2audio_replace_buffer(audio, allocation_needed);
	}

// Entire output is in buffer so don't do anything.
	if(start_position >= audio->pcm_position && start_position < audio->pcm_position + audio->pcm_size &&
		start_position + len <= audio->pcm_size)
	{
		;
	}
	else
// Output starts in buffer but ends later so slide it back.
	if(start_position <= audio->pcm_position + audio->pcm_size &&
		start_position >= audio->pcm_position)
	{
		for(i = 0, j = start_position - audio->pcm_position;
			j < audio->pcm_size;
			j++)
		{
			audio->pcm_sample[i++] = audio->pcm_sample[j * audio->stereo];
			if(audio->stereo > 1) audio->pcm_sample[i++] = audio->pcm_sample[j * audio->stereo + 1];
		}
		audio->pcm_point = i;
		audio->pcm_size -= start_position - audio->pcm_position;
		audio->pcm_position = start_position;
	}
	else
	{
// Output is outside buffer completely
		result = mpeg2audio_seek(audio, start_position);
	}

// Read packets until the buffer is full.
	while(audio->pcm_size + audio->pcm_position < start_position + len && 
		!result)
	{
		result = mpeg2audio_read_frame(audio);
		audio->pcm_size = audio->pcm_point / audio->stereo;
	}

// Copy the buffer to the output
	if(output_f)
	{
		for(i = 0, j = (start_position - audio->pcm_position) * audio->stereo + channel; 
			i < len && j < audio->pcm_size * audio->stereo; 
			i++, j += audio->stereo)
		{
			output_f[i] = audio->pcm_sample[j];
		}
		for( ; i < len; i++)
		{
			output_f[i] = 0;
		}
	}
	else
	if(output_i)
	{
		int sample;
		for(i = 0, j = (start_position - audio->pcm_position) * audio->stereo + channel; 
			i < len && j < audio->pcm_size * audio->stereo; 
			i++, j += audio->stereo)
		{
			sample = (int)(audio->pcm_sample[j] * 32767);
			if(sample > 32767) sample = 32767;
			else 
			if(sample < -32768) sample = -32768;
			
			output_i[i] = sample;
		}
		for( ; i < len; i++)
		{
			output_i[i] = 0;
		}
	}

	if(audio->pcm_point > 0)
		return 0;
	else
		return result;
}
