#include "audiodevice.h"
#include "buttonbar.h"
#include "defaults.h"
#include "mainwindow.h"
#include "mainwindowgui.h"
#include "modules.h"
#include "patchbay.h"
#include "playbackcursor.h"
#include "playbackengine.h"
#include "renderengine.h"
#include "trackcanvas.h"
#include "videodevice.h"
#include "vrender.h"

PlaybackEngine::PlaybackEngine(MainWindow *mwindow)
 : Thread()
{
	this->mwindow = mwindow;
// reset parameters here that you want to be persistant between playbacks
	actual_frame_rate = -1;
	reset_parameters();
}

PlaybackEngine::~PlaybackEngine()
{
}


int PlaybackEngine::reset_parameters()
{
// called before every playback
	is_playing_back = 0;
	follow_loop = 0;
	speed = 1;
	reverse = 0;
	cursor = 0;
	last_position = 0;
	playback_start = playback_end = 0;
	infinite = 0;
	use_buttons = 0;
	audio = 0;
	video = 0;
	shared_audio = 0;
}


int PlaybackEngine::load_defaults(Defaults *defaults)
{
	sprintf(playback_device, "/dev/dsp");
	defaults->get("PLAYBACKDEVICE", playback_device);
	playback_bitrate = defaults->get("PLAYBACKBITRATE", 16);
	playback_channels = defaults->get("PLAYBACKCHANNELS", 2);
	view_follows_playback = defaults->get("VIEWFOLLOWSPLAYBACK", 1);
	real_time_playback = defaults->get("PLAYBACKREALTIME", 0);
	scrub_speed = defaults->get("SCRUBSPEED", (float)2);
	playback_software_timer = defaults->get("PLAYBACKSOFTWARETIMER", 1);
	audio_read_length = defaults->get("PLAYBACKREADBUFFER", 65536);
	amodule_render_fragment = defaults->get("AMODULEFRAGMENT", 4096);
	playback_buffer = defaults->get("PLAYBACKBUFFER", 4096);
	video_every_frame = defaults->get("VIDEOEVERYFRAME", 0);
	sprintf(video_playback_device, "");
	defaults->get("VIDEOPLAYBACKDEVICE", video_playback_device);
	video_read_length = 1;
	video_output_length = 1;
	actual_frame_rate = defaults->get("ACTUALFRAMERATE", (float)-1);
}

int PlaybackEngine::save_defaults(Defaults *defaults)
{
	defaults->update("PLAYBACKDEVICE", playback_device);
	defaults->update("PLAYBACKCHANNELS", playback_channels);
	defaults->update("PLAYBACKBITRATE", playback_bitrate);
	defaults->update("VIEWFOLLOWSPLAYBACK", view_follows_playback);
	defaults->update("PLAYBACKREALTIME", real_time_playback);
	defaults->update("SCRUBSPEED", scrub_speed);
	defaults->update("PLAYBACKSOFTWARETIMER", playback_software_timer);
	defaults->update("PLAYBACKREADBUFFER", audio_read_length);
	defaults->update("AMODULEFRAGMENT", amodule_render_fragment);
	defaults->update("PLAYBACKBUFFER", playback_buffer);
	defaults->update("VIDEOEVERYFRAME", video_every_frame);
	defaults->update("VIDEOPLAYBACKDEVICE", video_playback_device);
	defaults->update("ACTUALFRAMERATE", actual_frame_rate);
}

int PlaybackEngine::get_device_options(
                char *playback_device, 
				int *playback_bitrate, 
				int *playback_channels, 
				int *view_follows_playback, 
				long *playback_buffer, 
				int *real_time_playback, 
				int *playback_software_timer,
				long *read_length,
				long *module_render_fragment,
				int *video_every_frame,
				int *video_read_length,
				int *video_output_length,
				float *actual_frame_rate,
				char *video_playback_device)
{
	strcpy(playback_device, this->playback_device);
	strcpy(video_playback_device, this->video_playback_device);
	*playback_buffer = this->playback_buffer;
	*playback_bitrate = this->playback_bitrate;
	*playback_channels = this->playback_channels;
	*view_follows_playback = this->view_follows_playback;
	*real_time_playback = this->real_time_playback;
	*playback_software_timer = this->playback_software_timer;
	*read_length = this->audio_read_length;
	*module_render_fragment = this->amodule_render_fragment;
	*video_every_frame = this->video_every_frame;
	*video_read_length = this->video_read_length;
	*video_output_length = this->video_output_length;
	*actual_frame_rate = this->actual_frame_rate;
}

int PlaybackEngine::update_device_options(
								char *playback_device, 
								int playback_bitrate, 
								int playback_channels, 
								int view_follows_playback, 
								long playback_buffer, 
								int real_time_playback,
								int playback_software_timer,
								long read_length,
								long module_render_fragment,
								int video_every_frame,
								int video_read_length,
								int video_output_length,
								float actual_frame_rate,
								char *video_playback_device)
{
	strcpy(this->playback_device, playback_device);
	strcpy(this->video_playback_device, video_playback_device);
	if(playback_buffer > 0) this->playback_buffer = playback_buffer;
	if(playback_bitrate > 0 && playback_bitrate < 32) this->playback_bitrate = playback_bitrate;
	if(playback_channels > 0 && playback_channels < 10) this->playback_channels = playback_channels;
	this->view_follows_playback = view_follows_playback;
	this->real_time_playback = real_time_playback;
	this->playback_software_timer = playback_software_timer;
	this->audio_read_length = read_length;
	this->amodule_render_fragment = module_render_fragment;
	this->video_every_frame = video_every_frame;
	this->actual_frame_rate = actual_frame_rate;
}


// ===================================== initialization


int PlaybackEngine::init_parameters()
{
	is_playing_back = 1;
	update_button = 1;
	correction_factor = 0;

// correct playback buffer sizes
	playback_buffer = output_length = amodule_render_fragment;

// get maximum actual buffer size written to device plus padding
	if(speed != 1) output_length = (long)(output_length / speed) + 16;   
	if(output_length < playback_buffer) output_length = playback_buffer;

// samples to read at a time is a multiple of the playback buffer greater than read_length
	input_length = playback_buffer;
	while(input_length < audio_read_length) input_length += playback_buffer;
}

int PlaybackEngine::init_cursor()
{
// initialize playback cursor
	if(mwindow->gui)
	{
		cursor = new PlaybackCursor(mwindow);
		cursor->arm_playback(last_position, view_follows_playback, reverse, follow_loop);
	}
}

int PlaybackEngine::init_audio_device()
{
// open audio device
	if(!shared_audio)
	{
		audio = new AudioDevice;
		audio->open_output(playback_device, playback_bitrate, mwindow->sample_rate, playback_channels, output_length);
		if(playback_software_timer) audio->set_software_positioning();
	}
}

int PlaybackEngine::init_video_device()
{
// open audio device
	if(actual_frame_rate < 0) actual_frame_rate = mwindow->frame_rate;
	video = new VideoDevice(mwindow);
	video->open_output(mwindow->video_window, mwindow->frame_rate, mwindow->output_w, mwindow->output_h, video_output_length);
}


int PlaybackEngine::set_range(long start,
					long end,
					long current,
					int reverse,
					float speed)
{
	this->playback_start = start;
	this->playback_end = end;
	this->last_position = current;
	this->reverse = reverse;
	this->speed = speed;
	if(speed == FRAME_SPEED)
	{
// Fix positions for frame advance
		long frame = (long)(((float)(last_position + 1) / mwindow->sample_rate) * mwindow->frame_rate);
		last_position = (long)(((float)frame / mwindow->frame_rate) * mwindow->sample_rate);

		if(reverse)
		{
			playback_end = last_position;
			playback_start = (long)(((float)(frame - 1) / mwindow->frame_rate) * mwindow->sample_rate);
		}
		else
		{
			playback_start = last_position;
			playback_end = (long)(((float)(frame + 1) / mwindow->frame_rate) * mwindow->sample_rate);
		}
	}
}


int PlaybackEngine::arm_playback(int follow_loop, 
					int use_buttons, 
					int infinite, 
					AudioDevice *audio)
{
	this->follow_loop = follow_loop;
	this->infinite = infinite;
	this->use_buttons = use_buttons;
	this->audio = audio;
	is_playing_back = 1;   // Armed but no thread
	init_cursor();  // depends on values
}

int PlaybackEngine::start_playback()
{
// Originally in arm_playback
	int need_audio = mwindow->patches->total_playable_atracks();
	int need_video = mwindow->patches->total_playable_vtracks();
	if(speed == FRAME_SPEED) 
	{
		need_audio = 0;     // No audio during frame advance
	}
	if(!need_audio) audio = 0;    // Defeat a setting from arm_playback

// need to set these here since mwindow isn't created on time
	shared_audio = audio ? 1 : 0;

	init_parameters();
	if(need_audio) init_audio_device();
	if(need_video) init_video_device();

// initialize the render engine
	render_engine = new RenderEngine(mwindow, 
									this->audio, 
									video, 
									this); // resets on construction

	render_engine->arm_playback_common(playback_start,
						playback_end,
						last_position,
						reverse,
						speed,
						follow_loop,
						infinite,
						real_time_playback);

	if(need_audio)
	{
		render_engine->arm_playback_audio(input_length, 
						amodule_render_fragment, 
						playback_buffer, 
						output_length, 
						mwindow->output_channels);
	}

	if(need_video)
	{
		render_engine->arm_playback_video(video_every_frame, 
						video_read_length, 
						video_output_length, 
						mwindow->track_w,
						mwindow->track_h,
						mwindow->output_w,
						mwindow->output_h);
	}

	render_engine->wait_for_startup();


// Originally in start_playback
	is_playing_back = 2;
	complete.lock();
	timer.update();
	if(video) render_engine->start_video();	// start the video rendering
	if(audio) audio->start_playback();	// start the audio device playing back
	loop_adjustment = 0;
	if(cursor) cursor->start_playback();   // start the cursor and meters moving
	startup_lock.lock();
	start();
	wait_for_startup();

// Handle frame advance
	if(speed == FRAME_SPEED)
	{
		float frame;
// For frame advance only the frame under the cursor is played back then playback is paused
// before returning to the buttonbar.
		mwindow->gui->unlock_window();
		wait_for_completion();

// Determine next playback range
		frame = (float)(playback_start + 1) / mwindow->sample_rate * mwindow->frame_rate;
		if(reverse)
		{
			playback_start = (long)((frame - 1) / mwindow->frame_rate * mwindow->sample_rate);
			playback_end = (long)(frame / mwindow->frame_rate * mwindow->sample_rate);
			last_position = playback_end;
		}
		else
		{
			playback_start = (long)((frame + 1) / mwindow->frame_rate * mwindow->sample_rate);
			playback_end = (long)((frame + 2) / mwindow->frame_rate * mwindow->sample_rate);
			last_position = playback_start;
		}

// Restart the playback engine.
		arm_playback(follow_loop, 
					use_buttons, 
					infinite, 
					audio);

		mwindow->gui->lock_window();
		if(use_buttons) mwindow->gui->buttonbar->pause_transport();
	}
}

int PlaybackEngine::stop_playback(int update_button)
{
	long new_position = 0;
	this->update_button = update_button;

	if(is_playing_back == 2)
	{
// output device is running
		new_position = get_position(0);

// interrupt audio driver forcefully
		if(audio) audio->interrupt_playback();
		if(video) video->interrupt_playback();
// wait for finish
		wait_for_completion();
	}
	else
	if(is_playing_back == 1)
	{
// armed but no thread
		new_position = last_position;
		if(cursor) cursor->stop_playback();
		cleanup();
		reset_buttons();
	}

	is_playing_back = 0;
	if(reverse) playback_end = new_position; else playback_start = new_position;
	last_position = new_position;
}

int PlaybackEngine::start_reconfigure()
{
	reconfigure_status = is_playing_back;
	if(is_playing_back) stop_playback(0);
}

int PlaybackEngine::stop_reconfigure()
{
	if(reconfigure_status)
	{
		arm_playback(follow_loop,
					use_buttons,
					infinite,
					shared_audio ? audio : 0);

		if(reconfigure_status == 2) start_playback(); // don't start if originally paused
		reconfigure_status = 0;
	}
}

int PlaybackEngine::reset_buttons()
{
	if(mwindow->gui && update_button) 
	{
		mwindow->gui->buttonbar->reset_transport();
	}
}


long PlaybackEngine::absolute_position(int sync_time)
{
	long result;

	switch(is_playing_back)
	{
		case 2:
// Playing back now.
// try getting from audio device first
			if(audio)
			{
				result = audio->current_position();
			}
			else
// try getting from the video thread only if drawing cursor
			if(video && !sync_time)
			{
				result = (long)((float)render_engine->vrender->absolute_frame / mwindow->frame_rate * mwindow->sample_rate / speed);
			}
			else
// use elapsed time for synchronization
			{
				result = timer.get_scaled_difference(mwindow->sample_rate);
			}
			break;

		case 1:
// Paused
			result = last_position;
			break;
		
		default:
			result = 0;
	}

	return result;
}

long PlaybackEngine::get_position(int sync_time)
{
	long result;

	switch(is_playing_back)
	{
		case 2:
// playing back
			result = absolute_position(sync_time);
// adjust for speed
			result = (long)(result * speed);

// adjust direction and initial position
			if(reverse)
			{
				result = playback_end - result;
				result += loop_adjustment;

// adjust for looping
				while(mwindow->loop_playback && follow_loop && 
					result < mwindow->loop_start)
				{
					result += mwindow->loop_end - mwindow->loop_start;
					loop_adjustment += mwindow->loop_end - mwindow->loop_start;
				}
			}
			else
			{
				result += playback_start;
				result -= loop_adjustment;
				
				while(mwindow->loop_playback && follow_loop && 
					result > mwindow->loop_end)
				{
					result -= mwindow->loop_end - mwindow->loop_start;
					loop_adjustment += mwindow->loop_end - mwindow->loop_start;
				}
			}
			break;

		case 1:
// paused
			result = last_position;
			break;

		default:
// no value
			result = -1;
			break;
	}

	return result;
}


int PlaybackEngine::move_right(long distance) { mwindow->move_right(distance); }


int PlaybackEngine::cleanup()
{
	is_playing_back = 0;
	if(!shared_audio && audio)
	{
		delete audio;
		audio = 0;
	}
	
	if(video)
	{
		delete video;
		video = 0;
	}

	if(render_engine) delete render_engine;
	if(cursor) delete cursor;
	render_engine = 0;
	cursor = 0;
}

int PlaybackEngine::wait_for_startup()
{
	if(audio) audio->wait_for_startup();
	if(cursor) cursor->wait_for_startup();

	startup_lock.lock();
	startup_lock.unlock();
}

int PlaybackEngine::wait_for_completion()
{
// must use mutex not join since playback engine also terminates itself
	complete.lock();
	complete.unlock();
}





void PlaybackEngine::run()
{
	startup_lock.unlock();
	render_engine->wait_for_completion();


	reset_buttons();      // change the playbutton to play if there is one
	if(cursor) cursor->stop_playback();        // kill cursor

	if(!shared_audio && audio) // delete audio objects if not shared
	{
		audio->close_all();
	}
	
	if(video) video->close_all();

// delete more objects
	cleanup();                  
	complete.unlock();             // signal blocking threads to continue
}
