#include "assets.h"
#include "cache.h"
#include "console.h"
#include "filesystem.h"
#include "indexprefs.h"
#include "levelwindow.h"
#include "mainwindow.h"
#include "mainwindowgui.h"
#include "modules.h"
#include "playbackengine.h"
#include "playbackprefs.h"
#include "pluginprefs.h"
#include "preferences.h"
#include "recordprefs.h"
#include "viewprefs.h"
#include "videoprefs.h"

Preferences::Preferences(MainWindow *mwindow)
 : BC_MenuItem("Preferences...", ""), Thread() 
{ this->mwindow = mwindow; default_dialog = 0; }
 
int Preferences::handle_event() { start(); }

void Preferences::run()
{
	int need_new_indexes;

	mwindow->gui->disable_window();
	{
		PreferencesWindow preferences;

		preferences.create_objects(mwindow, this);
		preferences.run_window();
		need_new_indexes = preferences.need_new_indexes;
	}

	if(need_new_indexes)
	{           // indexes were deleted
		mwindow->lock_resize();
// ==================================== create new index files
		mwindow->assets->build_indexes();
// ==================================== draw
		mwindow->draw();
		mwindow->unlock_resize();
	}

	mwindow->gui->enable_window();
}

PreferencesWindow::PreferencesWindow(char *display)
 : BC_Window(display, MEGREY, ICONNAME ": Preferences", PREFSUBWINDOW_W + 160, PREFSUBWINDOW_H + 80, 0, 0)
{ need_new_indexes = 0; }

PreferencesWindow::~PreferencesWindow()
{
	delete_current_dialog();
	delete categorywindow;
	delete commandwindow;
}

int PreferencesWindow::create_objects(MainWindow *mwindow, Preferences *thread)
{
	this->mwindow = mwindow;
	this->thread = thread;
// ==================================== copy values
// ==================================== playback

	mwindow->playback_engine->get_device_options(playback_device, 
		&playback_bitrate, 
		&playback_channels, 
		&view_follows_playback, 
		&playback_buffer, 
		&real_time_playback, 
		&playback_software_timer,
		&playback_read_length,
		&playback_module_fragment,
		&video_every_frame,
		&video_read_length, 
		&video_output_length,
		&actual_frame_rate,
		video_playback_device);

	playback_preload = mwindow->playback_preload;
	test_playback_edits = mwindow->test_playback_edits;

// ===================================== video
	video_floatingpoint = mwindow->video_floatingpoint;
	video_use_alpha = mwindow->video_use_alpha;
	video_interpolate = mwindow->video_interpolate;
	video_driver = mwindow->video_driver;
	video_driver = mwindow->video_driver;
	video_write_length = mwindow->video_write_length;
	video_capture_length = mwindow->video_capture_length;

// ===================================== record
	strcpy(record_device, mwindow->record_device);
	strcpy(video_record_device, mwindow->video_record_device);
	record_bitrate = mwindow->record_bitrate;
	record_channels = mwindow->record_channels;
	strcpy(duplex_device, mwindow->duplex_device);
	duplex_bitrate = mwindow->duplex_bitrate;
	duplex_channels = mwindow->duplex_channels;	
	enable_duplex = mwindow->enable_duplex;
	record_write_length = mwindow->record_write_length;
	real_time_record = mwindow->real_time_record;

// ===================================== performance

	strcpy(index_directory, mwindow->index_directory);
	index_size = mwindow->index_size;
	index_count = mwindow->index_count;
	cache_size = mwindow->cache->recommended_bytes;
	smp = mwindow->smp;

// ===================================== interface

	time_format = mwindow->time_format;
	min_record_db = mwindow->min_meter_db;
	vu_format = mwindow->meter_format;
	edit_handle_mode[0] = mwindow->edit_handle_mode[0];
	edit_handle_mode[1] = mwindow->edit_handle_mode[1];
	edit_handle_mode[2] = mwindow->edit_handle_mode[2];
	frames_per_foot = mwindow->frames_per_foot;

// ===================================== plugins

	strcpy(global_plugin_dir, mwindow->global_plugin_dir);
	strcpy(local_plugin_dir, mwindow->local_plugin_dir);

	add_subwindow(categorywindow = new CategoryWindow(this, this));
	add_subwindow(commandwindow = new CommandWindow(this, this));
	
// ===================================== default preferences dialog

	set_current_dialog(thread->default_dialog, 1);
}

int PreferencesWindow::delete_current_dialog()
{
	switch(thread->default_dialog)
	{
		case 0:
			delete playbackprefs;
			break;
	
		case 1:
			delete recordprefs;
			break;
	
		case 2:
			delete indexprefs;
			break;
	
		case 3:
			delete viewprefs;
			break;
	
		case 4:
			delete pluginprefs;
			break;
	
		case 5:
			delete videoprefs;
			break;
	
		case 6:
			break;
	}
}

int PreferencesWindow::set_current_dialog(int number, int first_time)
{
	if(!first_time) delete_current_dialog();
	thread->default_dialog = number;
	switch(number)
	{
		case 0:
			add_subwindow(playbackprefs = new PlaybackPrefs(this));
			break;
	
		case 1:
			add_subwindow(recordprefs = new RecordPrefs(this));
			break;
	
		case 2:
			add_subwindow(indexprefs = new IndexPrefs(this));
			break;
	
		case 3:
			add_subwindow(viewprefs = new ViewPrefs(this));
			break;
	
		case 4:
			add_subwindow(pluginprefs = new PluginPrefs(this));
			break;
	
		case 5:
			add_subwindow(videoprefs = new VideoPrefs(this));
			break;
	
		case 6:
			break;
	}
}

// ================================== save values
int PreferencesWindow::save_values()
{
	if(mwindow->gui) mwindow->gui->lock_window();
// ==================================== playback
	mwindow->playback_engine->update_device_options(playback_device, 
		playback_bitrate, 
		playback_channels, 
		view_follows_playback, 
		playback_buffer, 
		real_time_playback, 
		playback_software_timer,
		playback_read_length,
		playback_module_fragment,
		video_every_frame,
		video_read_length,
		video_output_length,
		actual_frame_rate,
		video_playback_device);

	mwindow->playback_preload = playback_preload;
	mwindow->test_playback_edits = test_playback_edits;

// ===================================== video
	mwindow->video_floatingpoint = video_floatingpoint;
	mwindow->video_use_alpha = video_use_alpha;
	mwindow->video_interpolate = video_interpolate;
	mwindow->video_driver = video_driver;
	if(video_write_length < 1) video_write_length = 1;
	mwindow->video_write_length = video_write_length;
	if(video_capture_length < 1) video_capture_length = 1;
	mwindow->video_capture_length = video_capture_length;
	strcpy(mwindow->video_record_device, video_record_device);

// ==================================== record

	strcpy(mwindow->record_device, record_device);
	if(record_channels > 0 && record_channels < 10) mwindow->record_channels = record_channels;
	if(record_bitrate > 0 && record_bitrate < 32) mwindow->record_bitrate = record_bitrate;
	strcpy(mwindow->duplex_device, duplex_device);
	if(duplex_channels > 0 && duplex_channels < 10) mwindow->duplex_channels = duplex_channels;
	if(duplex_channels > 0 && duplex_channels < 32) mwindow->duplex_bitrate = duplex_bitrate;
	mwindow->enable_duplex = enable_duplex;
	mwindow->record_write_length = record_write_length;
	mwindow->real_time_record = real_time_record;

// ===================================== performance
	FileSystem directory;
	directory.complete_path(index_directory);
	directory.add_end_slash(index_directory);
	strcpy(mwindow->index_directory, index_directory);
	mwindow->index_size = index_size;
	mwindow->index_count = index_count;
	mwindow->cache->recommended_bytes = cache_size;
	mwindow->smp = smp;

// ===================================== interface
	mwindow->time_format = time_format;
	mwindow->edit_handle_mode[0] = edit_handle_mode[0];
	mwindow->edit_handle_mode[1] = edit_handle_mode[1];
	mwindow->edit_handle_mode[2] = edit_handle_mode[2];
	mwindow->frames_per_foot = frames_per_foot;

	if(min_record_db > -200 && min_record_db < 0 &&
		mwindow->min_meter_db != min_record_db) mwindow->min_meter_db = min_record_db;

	if(mwindow->meter_format != vu_format)
	{
	  mwindow->meter_format = vu_format;
	  mwindow->level_window->change_format();
	  mwindow->console->modules->change_format();
	}
	mwindow->redraw_time_dependancies();

// ===================================== plugins

	strcpy(mwindow->global_plugin_dir, global_plugin_dir);
	strcpy(mwindow->local_plugin_dir, local_plugin_dir);
	if(mwindow->gui) mwindow->gui->unlock_window();
}




// ============================== category window




CategoryWindow::CategoryWindow(BC_Window *window, PreferencesWindow *pwindow)
 : BC_SubWindow(10, 10, 130, PREFSUBWINDOW_H, MEGREY)
{ this->pwindow = pwindow; }

CategoryWindow::~CategoryWindow()
{
	delete playbackbutton;
	delete recordbutton;
	delete indexbutton;
	delete viewbutton;
	delete pluginbutton;
}

int CategoryWindow::create_objects()
{
	add_border(get_resources()->get_bg_shadow1(),
		get_resources()->get_bg_shadow2(),
		get_resources()->get_bg_color(),
		get_resources()->get_bg_light2(),
		get_resources()->get_bg_light1());
	add_tool(new BC_Title(5, 5, "Categories", LARGEFONT, RED));
	add_tool(playbackbutton = new PlaybackButton(pwindow));
	add_tool(recordbutton = new RecordButton(pwindow));
	add_tool(indexbutton = new IndexButton(pwindow));
	add_tool(viewbutton = new ViewButton(pwindow));
	add_tool(pluginbutton = new PluginPrefsButton(pwindow));
	add_tool(new VideoPrefsButton(pwindow));
}



// ================================= command window





CommandWindow::CommandWindow(BC_Window *window, PreferencesWindow *pwindow)
 : BC_SubWindow(10, PREFSUBWINDOW_H + 20, 580, 50, MEGREY)
{
	this->pwindow = pwindow;
}

CommandWindow::~CommandWindow()
{
	delete ok;
	delete cancel;
}

int CommandWindow::create_objects()
{
	add_border(get_resources()->get_bg_shadow1(),
		get_resources()->get_bg_shadow2(),
		get_resources()->get_bg_color(),
		get_resources()->get_bg_light2(),
		get_resources()->get_bg_light1());
	add_tool(ok = new OkButton(pwindow));
	add_tool(cancel = new PrefCancelButton(pwindow));
}

OkButton::OkButton(PreferencesWindow *pwindow)
 : BC_BigButton(10, 10, "OK")
{
	this->pwindow = pwindow;
}

OkButton::~OkButton()
{
}

int OkButton::handle_event()
{
	pwindow->save_values();
	pwindow->set_done(0);
}

int OkButton::keypress_event()
{
	if(get_keypress() == 13) { handle_event(); return 1; }
	return 0;
}

PrefCancelButton::PrefCancelButton(PreferencesWindow *pwindow)
 : BC_BigButton(300, 10, "Cancel")
{
	this->pwindow = pwindow;
}

PrefCancelButton::~PrefCancelButton()
{
}

int PrefCancelButton::handle_event()
{
	pwindow->set_done(1);
}

int PrefCancelButton::keypress_event()
{
	if(get_keypress() == ESC) { handle_event();  return 1; }
	return 0;
}






// ==================================== category buttons






PlaybackButton::PlaybackButton(PreferencesWindow *pwindow)
 : BC_BigButton(5, 40, "Playback")
{ this->pwindow = pwindow; }
PlaybackButton::~PlaybackButton() {}
int PlaybackButton::handle_event()
{ pwindow->set_current_dialog(0); }

RecordButton::RecordButton(PreferencesWindow *pwindow)
 : BC_BigButton(5, 70, "Record")
{ this->pwindow = pwindow; }
RecordButton::~RecordButton()
{}
int RecordButton::handle_event()
{ pwindow->set_current_dialog(1); }

IndexButton::IndexButton(PreferencesWindow *pwindow)
 : BC_BigButton(5, 100, "Performance")
{ this->pwindow = pwindow; }
IndexButton::~IndexButton()
{ }
int IndexButton::handle_event()
{ pwindow->set_current_dialog(2); }

ViewButton::ViewButton(PreferencesWindow *pwindow)
 : BC_BigButton(5, 130, "Interface")
{ this->pwindow = pwindow; }
ViewButton::~ViewButton()
{ }
int ViewButton::handle_event()
{ pwindow->set_current_dialog(3); }

PluginPrefsButton::PluginPrefsButton(PreferencesWindow *pwindow)
 : BC_BigButton(5, 160, "Plugins")
{ this->pwindow = pwindow; }
PluginPrefsButton::~PluginPrefsButton()
{ }
int PluginPrefsButton::handle_event()
{ pwindow->set_current_dialog(4); }

VideoPrefsButton::VideoPrefsButton(PreferencesWindow *pwindow)
 : BC_BigButton(5, 190, "Video")
{ this->pwindow = pwindow; }
VideoPrefsButton::~VideoPrefsButton()
{ }
int VideoPrefsButton::handle_event()
{ pwindow->set_current_dialog(5); }
