#include "assets.h"
#include "audiodevice.h"
#include "file.h"
#include "filehtal.h"
#include "mainwindow.h"
#include "patchbay.h"
#include "playbackengine.h"
#include "previewaudio.h"
#include "previewvideo.h"
#include "recconfirmdelete.h"
#include "record.h"
#include "recordaudio.h"
#include "recordengine.h"
#include "recordgui.h"
#include "recordlabel.h"
#include "recordvideo.h"
#include "recvideowindow.h"
#include "videodevice.h"

#include <ctype.h>

RecordEngine::RecordEngine(MainWindow *mwindow, 
			Record *record, 
			File *file, 
			Asset *asset, 
			RecordLabels *labels)
{
	this->mwindow = mwindow;
	this->record = record; 
	this->file = file; 
	this->labels = labels; 
	this->asset = asset;
	is_saving = 0;
	is_previewing = 0;
	is_duplexing = 0;
	is_monitoring = 0;
	prev_label = -1;
	next_label = -1;

	if(record->do_audio) 
		device = new AudioDevice; 
	else 
		device = 0;

	if(record->do_video) 
		vdevice = new VideoDevice(mwindow); 
	else 
		vdevice = 0;
}

RecordEngine::~RecordEngine()
{
	if(record->do_audio)
	{
		delete monitor_audio;
		delete record_audio;
		delete preview_audio;
		delete device;
	}

	if(record->do_video)
	{
		delete monitor_video;
		delete record_video;
		delete preview_video;
		delete vdevice;
	}
}

int RecordEngine::create_objects()
{
// put at end of file
	total_length = -1;
	if(record->do_audio) current_position = file->get_audio_length();
	else
	if(record->do_video) current_position = tosamples((float)(file->get_video_length(record->get_framerate())), record->get_samplerate(), record->get_framerate());

	file->seek_end();

// =============================== create monitor thread

	if(record->do_audio) monitor_audio = new RecordAudio(record, this, device, 0);
	if(record->do_video) monitor_video = new RecordVideo(record, this, vdevice, 0);

// =============================== create record thread

	if(record->do_audio) record_audio = new RecordAudio(record, this, device, 1);
	if(record->do_video) record_video = new RecordVideo(record, this, vdevice, 1);

// ================================ create preview thread

	if(record->do_audio) preview_audio = new PreviewAudio(record, this);
	if(record->do_video) preview_video = new PreviewVideo(record, this);

// ================================= create duplex thread

	duplex_thread = mwindow->playback_engine;

// initialize seek buttons
	jump_delay[0] = 100;
	jump_delay[1] = 50;
	jump_delay[2] = 25;
	jump_delay[3] = 10;
	jump_delay[4] = 5;

	current_jump_jumps[0] = 20;
	current_jump_jumps[1] = 40;
	current_jump_jumps[2] = 60;
	current_jump_jumps[3] = 80;
	current_jump_jumps[4] = 100;
}

int RecordEngine::run_script(FileHTAL *script)
{
	int result = 0, script_result = 0;
	char string[1024];

	while(!result && !script_result)
	{
		result = script->read_tag();

		if(!result)
		{
			if(script->tag.title_is("set_mode"))
			{
				set_record_mode(script->tag.get_property_text(0));
				mode_to_text(string, get_record_mode());
				gui->rec_mode_menu->update(string);
			}
			else
			if(script->tag.title_is("set_duration"))
			{
				record->set_loop_duration((long)record->get_samplerate() * script->tag.get_property_int(0));
				gui->update_duration_boxes();
			}
			else
			if(script->tag.title_is("start_recording"))
			{
				gui->unlock_window();
				start_saving();
				gui->lock_window();
			}
			else
			if(script->tag.title_is("set_monitor_video"))
			{
				set_monitor_video(script->tag.get_property_int(0));
				if(!script->tag.get_property_int(0) && record->video_window_open)
				{
					record->video_window_open = 0;
					gui->monitor_video_window->window->hide_window();
				}
			}
			else
			if(script->tag.title_is("set_monitor_audio"))
			{
				set_monitor_audio(script->tag.get_property_int(0));
			}
			else
			if(script->tag.title_is("quit_when_completed"))
			{
				if(record->do_audio)
					record_audio->quit_when_completed = 1;
				if(record->do_video)
					record_video->quit_when_completed = 1;
			}
			else
			if(script->tag.title_is("ok"))
			{
				script_result = 1;
			}
		}
	}
	return script_result;
}

// ============================================= accounting

long RecordEngine::get_dc_offset(int offset)
{
	return record->dc_offset[offset];
}

int RecordEngine::set_gui(RecordGUI *gui)
{
	this->gui = gui;
	update_position(current_position);
}

int RecordEngine::get_duplex_enable()
{
	return record->enable_duplex();
}



// ================================================ operations

int RecordEngine::start_monitor()
{
	monitor_timer.update();
	if(record->do_audio)
	{
		monitor_audio->set_parameters(file, 
									gui, 
									gui->meter, 
									get_in_length(), 
									record->get_out_length(),
									asset->channels, 
									record->dc_offset,
									0);
		monitor_audio->start_recording();
	}
	if(record->do_video)
	{
		monitor_video->set_parameters(file, gui, 0, 0);
		monitor_video->start_recording();
	}
	is_monitoring = 1;
}

int RecordEngine::stop_monitor()
{
	if(is_monitoring)
	{
		is_monitoring = 0;
		if(record->do_audio) monitor_audio->stop_recording();
		if(record->do_video) monitor_video->stop_recording();
	}
}

int RecordEngine::pause_monitor()
{
	if(is_monitoring)
	{
		is_monitoring = 0;
		if(record->do_audio) monitor_audio->pause_recording();
		if(record->do_video) monitor_video->pause_recording();
	}
}

int RecordEngine::resume_monitor(int audio, int video)
{
	monitor_timer.update();
	if(audio)
	{
		if(record->do_audio) monitor_audio->resume_recording();
	}

	if(video)
	{
		if(record->do_video) monitor_video->resume_recording();
	}
	is_monitoring = 1;
}



int RecordEngine::start_saving(int duplex)
{
	if(!is_saving)
	{
		pause_monitor();

		record_timer.update();
// start the duplex engine if necessary
		if(record->enable_duplex() && duplex)
		{
			start_duplex();
		}
		else
		{
			if(record->do_audio) open_audio_input();
			if(record->do_video) open_video_input();
		}

		if(record->do_audio)
		{
			record_audio->set_parameters(file,
									gui,
									gui->meter, 
									get_in_length(), 
									record->get_in_buffersize(), 
									asset->channels, 
									record->dc_offset, 
									record->get_realtime());
			record_audio->start_recording();
		}

		if(record->do_video)
		{
			record_video->set_parameters(file, 
									gui, 
									record->get_video_buffersize(), 
									record->get_realtime());
			record_video->start_recording();
		}
		is_saving = 1;
	}
}

int RecordEngine::stop_saving()
{
	if(is_saving)
	{
		is_saving = 0;
		if(record->do_audio) record_audio->stop_recording(); // automatically stops duplex here and resumes monitor
		if(record->do_video) record_video->stop_recording(); // automatically stops duplex here and resumes monitor
		if(record->do_audio) record_audio->wait_for_completion();
		if(record->do_video) record_video->wait_for_completion();
	}
}

int RecordEngine::start_duplex()
{
	if(!is_duplexing)
	{
		open_duplex();
		arm_duplex();

// Start the playback engine.
		duplex_thread->start_playback();
//printf("RecordEngine::start_duplex 2\n");
// OSS < 3.9 crashes if recording starts before playback
// OSS >= 3.9 crashes if playback starts before recording
		is_duplexing = 1;
	}
}

int RecordEngine::stop_duplex()
{
	if(is_duplexing)
	{
		is_duplexing = 0;
		duplex_thread->stop_playback(0);
// OSS can't resume recording if buffers underrun
// so stop playback first
	}
}

int RecordEngine::start_preview()
{
	if(!is_previewing)
	{
		stop_operation();
		pause_monitor();

		preview_timer.update();
		if(record->do_audio)
		{
			preview_audio->set_parameters(gui->meter, device, record->get_out_buffersize(), record->get_samplerate());
			preview_audio->start_preview(current_position, file);
		}

		if(record->do_video)
		{
			preview_video->set_parameters(/*vdevice, */gui, file);
			preview_video->start_preview(current_position);
		}

		is_previewing = 1;
	}
}

int RecordEngine::stop_preview()
{
	if(is_previewing)
	{
		is_previewing = 0;
		if(record->do_audio) preview_audio->stop_preview();
		if(record->do_video) preview_video->stop_preview();
// preview engine automatically triggers monitor when finished
	}
}

int RecordEngine::stop_operation()
{
// Resumes monitoring after stopping
	if(is_saving) stop_saving();
	else
	if(is_previewing) stop_preview();
}

int RecordEngine::open_audio_input()
{
	if(record->do_audio)
	{
		device->open_input(record->get_in_path(), 
				record->get_in_bits(), 
				record->get_samplerate(), 
				record->get_in_channels(), 
				get_in_length());

// calibrate dc offsets
		for(int i = 0; i < asset->channels; i++)
		{
			device->set_dc_offset(record->dc_offset[i], i);
		}
	}
}

int RecordEngine::open_video_input()
{
	if(record->do_video)
	{
		vdevice->open_input(record->get_video_inpath(), record->get_video_driver(), 
			record->frame_w, record->frame_h, 
			record->video_x, record->video_y, record->video_zoom,
			record->get_video_capturesize(),
			get_frame_rate());
		if(record->get_current_channel())
			vdevice->set_channel(record->get_current_channel());
		set_video_picture();
	}
}

int RecordEngine::set_video_picture()
{
	if(record->do_video && vdevice) vdevice->set_picture(record->video_brightness,
		record->video_hue,
		record->video_color,
		record->video_contrast,
		record->video_whiteness);
}

int RecordEngine::close_audio_input()
{
	if(record->do_audio)
	{
		device->close_all();
	}
}

int RecordEngine::close_video_input()
{
// The following interrupts any video device in case of a lockup
	if(record->do_video)
	{
		vdevice->close_all();
	}
}

int RecordEngine::open_audio_output()
{
	if(record->do_audio)
	{
		device->open_output(record->get_out_path(), record->get_out_bits(), record->get_samplerate(), record->get_out_channels(), record->get_out_length());
		if(record->get_software_positioning()) device->set_software_positioning();
	}
}

int RecordEngine::open_video_output()
{
// This is never used
	if(record->do_video)
	{
		//vdevice->open_output(gui->monitor_video_window, mwindow->frame_rate, mwindow->output_w, mwindow->output_h, mwindow->playback_engine->video_output_length);
	}
}

int RecordEngine::close_audio_output()
{
	if(record->do_audio)
	{
		device->close_all();
	}
}

int RecordEngine::close_video_output()
{
// This is never used
	if(record->do_video)
	{
		//vdevice->close_all();
	}
}

int RecordEngine::open_duplex()
{
// Do the audio
	int need_playback_audio = mwindow->patches->total_playable_atracks();

	if(record->do_audio)
	{
// Only open in duplex mode if playback has audio
		if(need_playback_audio)
		{
			if((!strcmp(record->get_in_path(), record->get_duplex_path())) &&
				(record->get_in_bits() == record->get_duplex_bits()) &&
				(record->get_in_channels() == record->get_duplex_channels()))
			{      // duplex device is identical to input device
				device->open_duplex(record->get_duplex_path(), record->get_duplex_bits(), record->get_samplerate(), record->get_duplex_channels(), get_in_length());
			}
			else
			{      // two seperate devices
				open_audio_input();

				device->open_output(record->get_duplex_path(), record->get_duplex_bits(), record->get_samplerate(), record->get_duplex_channels(), record->get_out_length());
			}
		}
		else
		{
// Open in standard recording mode.
			open_audio_input();
		}
	}

	if(record->do_video)
	{
		vdevice->open_input(record->get_video_inpath(), record->get_video_driver(),
			record->frame_w, record->frame_h,
			record->video_x, record->video_y, record->video_zoom,
			record->get_video_capturesize(),
			get_frame_rate());
		if(record->get_current_channel())
			vdevice->set_channel(record->get_current_channel());
		set_video_picture();
	}

	if(record->do_audio && record->get_software_positioning()) device->set_software_positioning();
}

int RecordEngine::arm_duplex()
{
	if(get_duplex_enable())
	{
		long start, end;

		record->get_duplex_range(&start, &end);
		duplex_thread->reset_parameters();
		duplex_thread->set_range(start + current_position, end, start + current_position, 0, 1);
		duplex_thread->arm_playback(0, 0, 1, device);
	}
}



int RecordEngine::lock_window()
{
	gui->lock_window();
}

int RecordEngine::unlock_window()
{
	gui->unlock_window();
}

int RecordEngine::update_position(long new_position)
{
	if(new_position < 0) new_position = 0;      // fread error in filepcm
	current_position = new_position;
	
	gui->update_position(new_position);

	if(new_position > total_length) 
	{
		total_length = new_position;
		gui->update_total_length(new_position);
	}
	
	if(prev_label != labels->get_prev_label(new_position))
	{
		prev_label = labels->get_prev_label(new_position);
		gui->update_prev_label(prev_label);
	}
	
	if(next_label != labels->get_next_label(new_position))
	{
		next_label = labels->get_next_label(new_position);

		gui->update_next_label(next_label);
	}
}

int RecordEngine::goto_prev_label()
{
	if(!is_saving)
	{
		stop_operation();
		long new_position;

		new_position = labels->goto_prev_label(current_position);
		if(new_position != -1)
		{
			if(record->do_audio) file->set_audio_position(new_position);
			if(record->do_video) file->set_video_position(toframes(new_position, record->get_samplerate(), record->get_framerate()), record->get_framerate());
			update_position(new_position);
		}
	}
}

int RecordEngine::goto_next_label()
{
	if(!is_saving)
	{
		stop_operation();
		long new_position;

		new_position = labels->goto_next_label(current_position);
		if(new_position != -1 && new_position <= total_length)
		{
			if(record->do_audio) file->set_audio_position(new_position);
			if(record->do_video) file->set_video_position(toframes(new_position, record->get_samplerate(), record->get_framerate()), record->get_framerate());
			update_position(new_position);
		}
	}
}

int RecordEngine::toggle_label()
{
	labels->toggle_label(current_position);
	update_position(current_position);
}

int RecordEngine::calibrate_dc_offset()
{
	if(record->do_audio)
		monitor_audio->get_dc_offset(record->dc_offset, gui->dc_offset_text);
}

int RecordEngine::calibrate_dc_offset(long new_value, int channel)
{
	if(record->do_audio)
	{
		monitor_audio->set_dc_offset(new_value, channel);
		record->dc_offset[channel] = new_value;
	}
}

int RecordEngine::reset_over()
{
}

int RecordEngine::set_done(int value)
{
	stop_operation();
	stop_monitor();
	gui->set_done(value);
}

int RecordEngine::start_over()
{
	if((record->do_audio && file->get_audio_length() > 0) ||
		(record->do_video && file->get_video_length(record->get_framerate()) > 0))
	{
		RecConfirmDelete dialog;
		dialog.create_objects("start over");
		int result = dialog.run_window();
		if(!result)
		{
			stop_operation();
// remove file
			file->close_file();
			remove(asset->path);

// reopen file
			file->set_processors(mwindow->smp ? 2: 1);
			file->set_preload(mwindow->playback_preload);
			file->try_to_open_file(mwindow->plugindb, asset, 1, 1);

// start the engine over
			labels->delete_new_labels();
			update_position(0);
			total_length = 0;
			gui->update_total_length(0);

			record->startsource_sample = 0;
			record->startsource_frame = 0;
		}
	}
}

int RecordEngine::change_channel(Channel *channel)
{
	if(record->do_video && vdevice) 
		return vdevice->set_channel(channel);
	else
		return 0;
}

ArrayList<char*>* RecordEngine::get_video_inputs() 
{ 
	if(record->do_video && vdevice) 
		return vdevice->get_inputs();
	else
		return 0;
}

int RecordEngine::get_vu_format() { return record->get_vu_format(); }
int RecordEngine::get_duplex_status() { return record->duplex_status; }
int RecordEngine::set_duplex_status(int duplex_status) { record->duplex_status = duplex_status; }
int RecordEngine::get_dither() { return record->dither * asset->bits; }
int RecordEngine::get_input_channels() { return asset->channels; }
int RecordEngine::get_format(char *string)
{
	File file;
	strcpy(string, file.formattostr(mwindow->plugindb, asset->format)); 
}
int RecordEngine::get_samplerate() { return asset->rate; }
int RecordEngine::get_bits() { return asset->bits; }
int RecordEngine::get_time_format() { return record->get_time_format(); }
float RecordEngine::get_frame_rate() { return record->get_frame_rate(); }
int RecordEngine::get_loop_hr() { return record->loop_duration / asset->rate / 3600; }
int RecordEngine::get_loop_min() { return record->loop_duration / asset->rate / 60 - (long)get_loop_hr() * 60; }
int RecordEngine::get_loop_sec() { return record->loop_duration / asset->rate - (long)get_loop_hr() * 3600 - (long)get_loop_min() * 60; }
long RecordEngine::get_loop_duration() { return record->loop_duration; }
float RecordEngine::get_min_db() { return record->get_min_db(); }
int RecordEngine::get_meter_over_hold(int divisions) { return divisions * 15; }
int RecordEngine::get_meter_peak_hold(int divisions) { return divisions * 2; }
int RecordEngine::get_meter_speed() { return record->get_meter_speed(); }
float RecordEngine::get_frames_per_foot() { return mwindow->frames_per_foot; }

int RecordEngine::set_monitor_video(int value)
{
	if(!record->monitor_video && value && !record->video_window_open)
	{
		gui->monitor_video_window->window->show_window();
	}
	if(gui->monitor_video_toggle && gui->monitor_video_toggle->get_value() != value) 
		gui->monitor_video_toggle->update(value);
	record->monitor_video = value;
	if(value) record->video_window_open = 1;
}

int RecordEngine::set_monitor_audio(int value)
{
	record->monitor_audio = value;
	if(gui->monitor_audio_toggle && gui->monitor_audio_toggle->get_value() != value) 
		gui->monitor_audio_toggle->update(value);
}

int RecordEngine::set_record_mode(char *text)
{
	record->record_mode = text_to_mode(text);
}

int RecordEngine::get_record_mode(char *text)
{
	mode_to_text(text, record->record_mode);
}

int RecordEngine::get_record_mode()
{
	return record->record_mode;
}

int RecordEngine::mode_to_text(char *string, int mode)
{
	switch(mode)
	{
		case 0:        sprintf(string, "Untimed");       break;
		case 1:        sprintf(string, "Timed");         break;
		case 2:        sprintf(string, "Loop");          break;
	}
}

int RecordEngine::text_to_mode(char *string)
{
	if(!strcasecmp(string, "untimed")) return 0;
	if(!strcasecmp(string, "timed"))   return 1;
	if(!strcasecmp(string, "loop"))    return 2;
}

long RecordEngine::get_current_delay()
{
	if(current_jump_jump > 0) current_jump_jump--;
	if(current_jump_jump == 0 && current_jump_delay < /*JUMP_DELAYS*/ 1)
	{
		current_jump_delay++;
		current_jump_jump = current_jump_jumps[current_jump_delay];
	}
	return jump_delay[current_jump_delay];
}

int RecordEngine::reset_current_delay()
{
	current_jump_delay = 0;
	current_jump_jump = current_jump_jumps[current_jump_delay];
}

int RecordEngine::set_loop_duration() 
{
	record->set_loop_duration((long)record->get_samplerate() * (atol(gui->loop_sec->get_text()) + atol(gui->loop_min->get_text()) * 60 + atol(gui->loop_hr->get_text()) * 3600)); 
}


// Remember to change meters if you change this.
// Return the size of the fragments to read from the audio device.
int RecordEngine::get_in_length() 
{
	long fragment_size = 1;
	while(fragment_size < asset->rate / record->get_meter_speed()) fragment_size *= 2;
	fragment_size /= 2;
	return fragment_size;
}

// Different absolute positions are defined for each operation so threads
// can end at different times without screwing up the frame synchronization.
long RecordEngine::absolute_monitor_position()
{
	if(is_monitoring)
	{
		if(record->do_audio)
		{
			return monitor_audio->absolute_position();
		}
		else
		{
			return (long)((float)monitor_timer.get_difference() / 1000 * record->get_samplerate());
		}
	}
	else
	return -1;
}

long RecordEngine::absolute_preview_position()
{
	if(is_previewing)
	{
		if(record->do_audio)
		{
			return preview_audio->absolute_position();
		}
		else
		{
			return (long)((float)preview_timer.get_difference() / 1000 * record->get_samplerate());
		}
	}
	else
	return -1;
}

long RecordEngine::absolute_record_position()
{
	if(is_saving)
	{
		if(record->do_audio)
		{
			return record_audio->absolute_position();
		}
		else
		{
			return (long)((float)record_timer.get_difference() / 1000 * record->get_samplerate());
		}
	}
	else
	return -1;
}

