#include "auto.h"
#include "autos.h"
#include "console.h"
#include "mainwindow.h"
#include "module.h"
#include "modules.h"
#include "patchbay.h"
#include "plugin.h"
#include "pluginbuffer.h"
#include "pluginserver.h"
#include "tracks.h"
#include "virtualnode.h"

VirtualNode::VirtualNode(MainWindow *mwindow, 
		Module *real_module, 
		Plugin *real_plugin,
		Patch *patch, 
		Track *track, 
		VirtualNode *parent_module, 
		PluginBuffer **input,
		PluginBuffer **output,
		long buffer_size,
		long fragment_size, 
		int double_buffers, 
		int input_is_master,
		int output_is_master,
		int in,
		int out)
{
	this->mwindow = mwindow;
	this->real_module = real_module;
	this->real_plugin = real_plugin;
	this->patch = patch;
	this->track = track;
	this->parent_module = parent_module;
	this->buffer_size = buffer_size;
	this->fragment_size = fragment_size;
	this->double_buffers = double_buffers;
// assume parent is supplying buffers for now
	for(int i = 0; i < double_buffers; i++)
	{
		this->output[i] = output[i];
		this->input[i] = input[i];
	}

	shared_input = 1;
	this->input_is_master = input_is_master;
	shared_output = 1;
	this->output_is_master = output_is_master;
	this->in = in;
	this->out = out;
	for(int i = 0; i < PLUGINS; i++)
	{
		virtual_plugins[i] = 0;
	}

	render_count = 0;
	plugin_server = 0;
	plugin_type = 0;
	waiting_real_plugin = 0;
	plugin_buffer_number = 0;
	fade_before = fade_after = 0;
	mute_before = mute_after = 0;
}

VirtualNode::~VirtualNode()
{
	for(int i = 0; i < PLUGINS; i++)
	{
		if(virtual_plugins[i]) delete virtual_plugins[i];
	}
}

int VirtualNode::expand()
{
	expand_buffers();

// module needs to know where the input data for the next process is
	data_in_input = 1;
//printf("VirtualNode::expand 1\n");
	if(real_module)
	{
// expand the plugins for a module
		for(int i = 0; i < PLUGINS; i++)
		{
			virtual_plugins[i] = 0;
			if(real_module->plugins[i]->on)
			{
				int plugin_type = real_module->plugins[i]->plugin_type;
				if(plugin_type == 3)
				{
// plugin is a module
					attach_virtual_module(i);
				}
				else
				if(plugin_type)
				{
// plugin is a plugin
					attach_virtual_plugin(i);
				}
			}
		}
	}
	else
	if(real_plugin)
	{
// attach to a real plugin for a plugin
// plugin always takes data from input to output
		expand_as_plugin();
	}
//printf("VirtualNode::expand 2\n");
}

int VirtualNode::expand_buffers()
{
// get new buffers where needed
	if(!out)
	{
// create shared memory for the output since parent doesn't want it
		create_output_buffer();
		shared_output = 0;
		output_is_master = 0;
	}

	if(!in)
	{
// create shared memory for null input since parent isn't giving it
		create_input_buffer();
		shared_input = 0;
		input_is_master = 0;
	}

// give the derived node its data types
	create_buffer_ptrs();
}


int VirtualNode::attach_virtual_module(int plugin_number)
{
	int real_module_number = real_module->plugins[plugin_number]->shared_module_location.module;
	Module *plugin_module = mwindow->console->modules->module_number(real_module_number);
	Plugin *real_plugin = real_module->plugins[plugin_number];
	Patch *patch = mwindow->patches->number(real_module_number);
	Track *track = mwindow->tracks->number(real_module_number);

	virtual_plugins[plugin_number]
		 = create_module(real_plugin, plugin_module, parent_module, patch, track);
	virtual_plugins[plugin_number]->expand();
	
// working data is now in output
	if(real_plugin->out) data_in_input = 0;
}

int VirtualNode::attach_virtual_plugin(int plugin_number)
{
	Plugin *real_plugin = real_module->plugins[plugin_number];
	virtual_plugins[plugin_number] = create_plugin(real_plugin);
	virtual_plugins[plugin_number]->expand();
	
// working data is now in output
	if(real_plugin->out) data_in_input = 0;
}

int VirtualNode::expand_as_plugin()
{
	plugin_type = real_plugin->plugin_type;

	if(plugin_type == 2)
	{
// attached to a shared plugin
// get the real plugin it's attached to
// redirect the real_plugin to the shared plugin
		int real_module_number = real_plugin->shared_plugin_location.module;
		int real_plugin_number = real_plugin->shared_plugin_location.plugin;
		Module *real_module = mwindow->console->modules->module_number(real_module_number);
		real_plugin = real_module->plugins[real_plugin_number];
	}

// Add to real plugin's list of virtual plugins for configuration updates
// and plugin_server initializations.
// Input and output are taken care of when the parent module creates this plugin.
// Get number for passing to render.
	plugin_buffer_number = real_plugin->attach_virtual_plugin(this, 
							input, 
							output, 
							input_is_master,
							output_is_master,
							double_buffers,
							input_is_master ? buffer_size : fragment_size,
							output_is_master ? buffer_size : fragment_size,
							fragment_size);
}

int VirtualNode::sort(ArrayList<VirtualNode*>*render_list)
{
	int result = 0, total_result = 0;

	if(real_module)
	{
		sort_as_module(render_list, result, total_result);
	}
	else
	{
		sort_as_plugin(render_list, result, total_result);
	}

	if(!result && total_result) result = total_result;
// if a plugin that wasn't patched out couldn't be rendered, try again

	return result;
}

int VirtualNode::sort_as_module(ArrayList<VirtualNode*>*render_list, int &result, int &total_result)
{
	for(int i = 0; i < PLUGINS && !result; i++)
	{
// stop when rendering can't continue without another higher level module
		if(virtual_plugins[i])
		{
			result = virtual_plugins[i]->sort(render_list);

			if(result && !virtual_plugins[i]->out)
			{
				total_result = 1;
				result = 0;
// couldn't render the last plugin but it wasn't patched out so continue to next plugin
			}
		}
	}

// all plugins rendered or not patched out
// render this module
	if(render_count == 0 && !result)
	{
		render_list->append(this);
		render_count++;
		result = 0;
	}
}

int VirtualNode::sort_as_plugin(ArrayList<VirtualNode*>*render_list, int &result, int &total_result)
{
// need to know if plugin requires all inputs to be armed before rendering
	int multichannel = 0;
	
	if(plugin_type == 1 || plugin_type == 2) multichannel = real_plugin->multichannel_shared();
	
	if(plugin_type == 1 && !multichannel)
	{
// unshared single channel plugin
// render now
		if(!render_count)
		{
			render_list->append(this);
			render_count++;
			result = 0;
		}
	}
	else
	if(plugin_type == 2 || multichannel)
	{
		if(!render_count)
		{
			if(!plugin_server->multichannel)
			{
// shared single channel plugin
// render now
				render_list->append(this);
				render_count++;
				result = 0;
			}
			else
			{
// shared multichannel plugin
// all buffers must be armed before rendering at the same time
				if(!waiting_real_plugin)
				{
					waiting_real_plugin = 1;
					result = real_plugin->sort(this);
					render_list->append(this);
					render_count++;
				}
				else
				{
// Assume it was rendered later in the first pass
					result = 0;
				}
			}
		}
	}
}

int VirtualNode::get_plugin_input(int &double_buffer_in, long &fragment_position_in,
							int &double_buffer_out, long &fragment_position_out,
							int double_buffer, long fragment_position)
{
	if(input_is_master)
	{
		double_buffer_in = double_buffer;
		fragment_position_in = fragment_position;
	}
	else
	{
		double_buffer_in = 0;
		fragment_position_in = 0;
	}

	if(output_is_master)
	{
		double_buffer_out = double_buffer;
		fragment_position_out = fragment_position;
	}
	else
	{
		double_buffer_out = 0;
		fragment_position_out = 0;
	}	
}

int VirtualNode::render_as_plugin(int double_buffer,
		long fragment_position,
		long fragment_len,
		long step)
{
// need numbers for actual buffers
	int double_buffer_in, double_buffer_out;
	long fragment_position_in, fragment_position_out;

	get_plugin_input(double_buffer_in, fragment_position_in,
					double_buffer_out, fragment_position_out,
					double_buffer, fragment_position);

	int multichannel = 0;

	if(plugin_type == 1 || plugin_type == 2) 
		multichannel = real_plugin->multichannel_shared();

	if(plugin_type == 1 && !multichannel)
	{
// standalone plugin
// plugins copy from their input to their output
		plugin_server->arm_buffer(plugin_buffer_number,
							fragment_position_in,
							fragment_position_out,
							double_buffer_in,
							double_buffer_out);
		plugin_server->process_realtime(fragment_len, step);
	}
	else
	if(plugin_type == 2 || multichannel)
	{
// render a shared plugin
// shared single channel plugin can be processed now
		if(!plugin_server->multichannel)
		{
			plugin_server->arm_buffer(plugin_buffer_number,
							fragment_position_in,
							fragment_position_out,
							double_buffer_in,
							double_buffer_out);
			plugin_server->process_realtime(fragment_len, step);
		}
		else
		{
// try multi channel plugin now
// flag real plugin by setting render count here
			render_count = 0;       
			real_plugin->render(double_buffer_in, double_buffer_out, 
							fragment_position_in, fragment_position_out,
							fragment_len, plugin_buffer_number, step);
		}
	}
}



int VirtualNode::longest_mute_fragment(int &automate, 
			long input_position,
			long &buffer_len,
			Autos *autos,
			Auto **before, 
			Auto **after)
{
	float mute_constant = 0;
	long mute_len;

	init_automation(automate, 
				mute_constant, 
				input_position,
				buffer_len,
				autos,
				before, 
				after);

// Muted all the way through
	if(!automate && mute_constant > 0) return 1;
// Playable all the way through
	if(!automate && mute_constant < 0) return -1;

// Changes somewhere
	current_auto = *before;
	if(!current_auto)
	{
		current_auto = autos->first;
	}

	if(current_auto)
	{
		mute_constant = current_auto->value;
		current_auto = current_auto->next;
	}
	else
	{
		mute_constant = -1;
		automate = 0;
	}

	if(current_auto) 
	{
		mute_len = reverse ? current_auto->position - input_start : 
			current_auto->position - input_start;
	}
	else
	{
		mute_len = buffer_len;
	}

	if(buffer_len > mute_len) buffer_len = mute_len;
	return (int)mute_constant;
}




int VirtualNode::init_automation(int &automate, 
				float &constant, 
				long input_position,
				long buffer_len,
				Autos *autos,
				Auto **before, 
				Auto **after)
{
	buffer_position = 0;

// set start and end boundaries for automation info
	input_start = reverse ? input_position - buffer_len : input_position;
	input_end = reverse ? input_position : input_position + buffer_len;

// test automation for constant value
// and set up *before and *after
	if(automate)
	{
		if(autos->automation_is_constant(input_start, input_end, before, after))
		{
			constant += autos->get_automation_constant(input_start, input_end, before, after);
			automate = 0;
		}
	}
	return automate;
}

int VirtualNode::init_slope(Autos *autos, Auto **before, Auto **after)
{
// apply automation
	current_auto = reverse ? *after : *before;
// no auto before start so use first auto in range
// already know there is an auto since automation isn't constant
	if(!current_auto)
	{
		current_auto = reverse ? autos->last : autos->first;
		slope_value = current_auto->value;
		slope_start = input_start;
		slope_position = 0;
	}
	else
	{
// otherwise get the first slope point and advance auto
		slope_value = current_auto->value;
		slope_start = current_auto->position;
		slope_position = reverse ? slope_start - input_end : input_start - slope_start;
		current_auto = reverse ? current_auto->previous : current_auto->next;
	}
}

int VirtualNode::get_slope(long buffer_len, long buffer_position)
{
// get the slope
	if(current_auto)
	{
		slope_end = reverse ? slope_start - current_auto->position : current_auto->position - slope_start;
		if(slope_end) 
			slope = (current_auto->value - slope_value) / slope_end;
		else
			slope = 0;
	}
	else
	{
		slope = 0;
		slope_end = buffer_len - buffer_position;
	}
}

int VirtualNode::advance_slope()
{
	if(current_auto) 
	{
		slope_start = current_auto->position;
		slope_value = current_auto->value;
		current_auto = reverse ? current_auto->previous : current_auto->next;
		slope_position = 0;
	}
}
