#ifndef VIDEODEVICE_H
#define VIDEODEVICE_H

#include "bcbase.h"
#include "bccapture.inc"
#include "mainwindow.inc"
#include "mutex.h"
#include "recvideowindow.inc"
#include "thread.h"
#include "timer.h"
#include "videowindow.inc"

#include <linux/fs.h>
#include <linux/kernel.h>
#include "videodev2.h"

// The keepalive thread runs continuously during recording.
// If the recording thread doesn't reset still_alive, failed is incremented.
// Failed is set to 0 if the recording thread resets still_alive.

// Number of seconds for keepalive to freak out
#define KEEPALIVE_DELAY 1

// Structure for video4linux2
struct tag_vimage
{
	struct v4l2_buffer vidbuf;
	char *data;
};

class VideoDevice;

class KeepaliveThread : public Thread
{
public:
	KeepaliveThread(VideoDevice *device);
	~KeepaliveThread();

	void run();
	int reset_keepalive();   // Call after frame capture to reset counter
	int get_failed();
	int start_keepalive();
	int stop();

	Timer timer;
	int still_alive;
	int failed;
	int interrupted;
	VideoDevice *device;
	Mutex startup_lock;
	int capturing;
};

class VideoDevice
{
public:
// Recording constructor
	VideoDevice();
// Playback constructor
	VideoDevice(MainWindow *mwindow);
	~VideoDevice();

	int close_all();
	int set_cloexec_flag(int desc, int value);

// ===================================== Recording
	int open_input(char *device, 
		int driver, 
		int frame_w, 
		int frame_h, 
		int input_x, 
		int input_y, 
		float input_z,
		int total_frames,
		float frame_rate);
	int start_recording();
	int stop_recording();
	int get_failed();  // Return 1 if capturing and locked up
// Schedule capture size to be changed.
	int set_translation(int input_x, int input_y, float input_z);
// Change the channel
	int set_channel(Channel *channel);
// Values fro -100 to 100
	int set_picture(int brightness, 
		int hue, 
		int color, 
		int contrast, 
		int whiteness);
	int set_mute(int muted);
// Set frames to clear after translation change.
	int set_latency_counter(int value);
	int capture_frame(int frame_number);  // Start the frame_number capturing
	int next_frame(int previous_frame);   // Next frame to capture
	int read_buffer(VFrame *frame);  // Read the next frame off the device
	int frame_to_vframe(VFrame *frame, unsigned char *input); // Translate the captured frame to a VFrame
	int reset_parameters();
	int update_translation();  // Change the capture size when ready
	int arm_buffer(VFrame *output);
	int get_output_w();     // get the output window width
	int get_output_h();     // get the output window height
	ArrayList<char *>* get_inputs();
	BC_Bitmap* get_bitmap();

// ================================== Playback
	int open_output(VideoWindow *video_out, 
					float rate, 
					int out_w, 
					int out_h, 
					long timed_length,
					int output_format = OUTPUT_RGB);
	int start_playback();
	int interrupt_playback();
	int wait_for_startup();
	int wait_for_completion();
	int output_visible();     // Whether the output is visible or not.
	int stop_playback();
	long current_position();     // last frame rendered
	int write_buffer(VFrame *output);   // absolute frame of last frame in buffer
// Playback Device can copy frames directly from the asset.
// This is for future expansion as all file formats can read RGB frames directly.
	int can_copy_from(Asset *asset, int output_w, int output_h); 

// Playback
	int interrupt;
	int output_format;   // Compression format required for direct frame copy

// Recording
	int thread_buffer_num;            // next buffer to be played
	float orate, irate;               // frame rates
	Timer buffer_timer;               // timer for displaying frames in the current buffer
	Timer rate_timer;                 // timer for getting frame rate
	int out_w, out_h;                 // size of output frame being fed to device during playback
	int in_w, in_h;                   // size of frame being read from device during recording
	int r, w;                         // modes
	long frame_delay;                 // time from start of previous frame to start of next frame in ms
	MainWindow *mwindow;              //
	BC_Bitmap *buffer;
	VideoWindow *video_out;            // window for output if playback or input if screen capture
	int buffer_w, buffer_h;   // dimensions of buffers written to window
	int is_playing_back, is_recording; // status of thread
	long timed_length; // number of frames used to get frame rate
	int driver;
	int total_frames;   // Number of frames to buffer for drivers that support it
	float frame_rate; // Frame rate to set in device
// Location of input frame in captured frame
	int frame_in_capture_x1, frame_in_capture_x2, frame_in_capture_y1, frame_in_capture_y2;
	int capture_in_frame_x1, capture_in_frame_x2, capture_in_frame_y1, capture_in_frame_y2;
// Size of raw captured frame
	int capture_w, capture_h;
	int input_x, input_y;
	float input_z;
// Captured frame size can only be changed when ready
	int new_input_x, new_input_y;
	float new_input_z;
	int frame_resized;
// When the frame is resized, need to clear all the buffer frames.
	int latency_counter;
	int capturing;
	int swap_bytes;
// All the input sources on the device
	ArrayList<char *> input_sources;

private:
	int init_video4linux();  // Set the frame size and parameters for video4linux capturing
	int init_video4linux2();  // Set the frame size and parameters for video4linux capturing
	int unmap_v4l_shmem();
	int unmap_v4l2_shmem();
	int close_v4l();
	int close_v4l2();
	int wait_v4l_frame();
	int wait_v4l2_frame();
	int read_v4l_frame(VFrame *frame);
	int read_v4l2_frame(VFrame *frame);
	int v4l1_get_inputs();       // Determine what inputs are available on the device
	int v4l1_set_channel(Channel *channel);
	int v4l1_get_norm(int norm);  // Get the #define from the norm
	int v4l1_set_mute(int muted);
	int v4l1_set_picture(int brightness, 
		int hue, 
		int color, 
		int contrast, 
		int whiteness);

// Video4Linux
	struct video_window window_params;
	struct video_picture picture_params;
	struct video_mbuf capture_params;  // Capture for Video4Linux

// Video4Linux 2
	struct v4l2_requestbuffers v4l2_buffers;
	struct v4l2_format v4l2_params;
	struct tag_vimage *v4l2_buffer_list;
	struct v4l2_streamparm v4l2_parm;

// Common
	int input_fd, output_fd;
	char *capture_buffer;      // sequentual capture buffers for v4l1 or read buffer for v4l2
	int capture_frame_number;    // number of frame to capture into
	int read_frame_number;       // number of the captured frame to read
	int shared_memory;   // Capturing directly to memory
	char input_device[1024]; // Device used for stopping and restarting during zoom changes
	KeepaliveThread *keepalive;

// Screen capture
	BC_Capture *capture_bitmap;
};



#endif
