#include "../mpeg2private.h"
#include "mpeg2video.h"

#include <stdlib.h>

int mpeg2_new_slice_buffer(mpeg2_slice_buffer_t *slice_buffer)
{
	pthread_mutexattr_t mutex_attr;

	slice_buffer->data = malloc(1024);
	slice_buffer->buffer_size = 0;
	slice_buffer->buffer_allocation = 1024;
	slice_buffer->current_position = 0;
	slice_buffer->bits_size = 0;
	slice_buffer->bits = 0;
	slice_buffer->done = 0;
	pthread_mutexattr_init(&mutex_attr);
	pthread_mutex_init(&(slice_buffer->completion_lock), &mutex_attr);
	return 0;
}

int mpeg2_delete_slice_buffer(mpeg2_slice_buffer_t *slice_buffer)
{
	free(slice_buffer->data);
	pthread_mutex_destroy(&(slice_buffer->completion_lock));
	return 0;
}

int mpeg2_expand_slice_buffer(mpeg2_slice_buffer_t *slice_buffer)
{
	int i;
	unsigned char *new_buffer = malloc(slice_buffer->buffer_allocation * 2);
	for(i = 0; i < slice_buffer->buffer_size; i++)
		new_buffer[i] = slice_buffer->data[i];
	free(slice_buffer->data);
	slice_buffer->data = new_buffer;
	slice_buffer->buffer_allocation *= 2;
	return 0;
}

void mpeg2slice_fillbits(mpeg2_slice_buffer_t *slice_buffer, int bits)
{
	while(slice_buffer->bits_size < bits)
	{
		slice_buffer->bits <<= 8;
		if(slice_buffer->current_position < slice_buffer->buffer_size)
			slice_buffer->bits |= slice_buffer->data[slice_buffer->current_position++];
		slice_buffer->bits_size += 8;
	}
}

unsigned int mpeg2slice_getbits(mpeg2_slice_buffer_t *slice_buffer, int bits)
{
	unsigned int result;

	mpeg2slice_fillbits(slice_buffer, bits);

	result = (slice_buffer->bits >> (slice_buffer->bits_size - bits)) & (0xffffffff >> (32 - bits));
	slice_buffer->bits_size -= bits;
	return result;
}

unsigned int mpeg2slice_showbits(mpeg2_slice_buffer_t *slice_buffer, int bits)
{
	unsigned int result;

	mpeg2slice_fillbits(slice_buffer, bits);

	result = (slice_buffer->bits >> (slice_buffer->bits_size - bits)) & (0xffffffff >> (32 - bits));
	return result;
}

unsigned int mpeg2slice_flushbits(mpeg2_slice_buffer_t *slice_buffer, int bits)
{
	mpeg2slice_fillbits(slice_buffer, bits);
	slice_buffer->bits_size -= bits;
	return 0;
}

int mpeg2_decode_slice(mpeg2_slice_t *slice)
{
	mpeg2video_t *video = slice->video;
	int comp;
	int MBA, MBAmax, MBAinc, mb_type, cbp, motion_type = 0, dct_type;
	int slice_vert_pos_ext;
	unsigned int code;
	int bx, by;
	int dc_dct_pred[3];
	int mv_count, mv_format, mvscale;
	int PMV[2][2][2], mv_field_sel[2][2];
	int dmv, dmvector[2];
	int qs;
	int stwtype, stwclass; 
	int SNRcbp;
	int i;

/* number of macroblocks per picture */
  	MBAmax = video->mb_width * video->mb_height;

/* field picture has half as many macroblocks as frame */
	if(video->pict_struct != FRAME_PICTURE)
	    MBAmax >>= 1; 

/* macroblock address */
  	MBA = 0; 
/* first macroblock in slice is not skipped */
  	MBAinc = 0;

  	slice->fault = 0;

	code = mpeg2slice_getbits(slice->slice_buffer, 32);
/* decode slice header (may change quant_scale) */
    slice_vert_pos_ext = mpeg2video_getslicehdr(slice, video, video->layerdata);

/* reset all DC coefficient and motion vector predictors */
    dc_dct_pred[0] = dc_dct_pred[1] = dc_dct_pred[2] = 0;
    PMV[0][0][0] = PMV[0][0][1] = PMV[1][0][0] = PMV[1][0][1] = 0;
    PMV[0][1][0] = PMV[0][1][1] = PMV[1][1][0] = PMV[1][1][1] = 0;

  	for(i = 0; 
		slice->slice_buffer->current_position < slice->slice_buffer->buffer_size; 
		i++)
	{
		if(MBAinc == 0)
		{
/* Done */
			if(!mpeg2slice_showbits(slice->slice_buffer, 23)) return 0;
/* decode macroblock address increment */
    		MBAinc = mpeg2video_get_macroblock_address(slice);
        	if(slice->fault) return 1;

    		if(i == 0)
			{
/* Get the MBA */
				MBA = ((slice_vert_pos_ext << 7) + (code & 255) - 1) * video->mb_width + MBAinc - 1;
/* first macroblock in slice: not skipped */
				MBAinc = 1;
			}
		}
//printf("debug slice 1 %x\n", mpeg2slice_showbits(slice->slice_buffer, 16));
        if(slice->fault) return 1;

    	if(MBA >= MBAmax)
		{
/* MBAinc points beyond picture dimensions */
      		fprintf(stderr, "mpeg2_decode_slice: too many macroblocks in picture\n");
      		return 1;
    	}

/* not skipped */
    	if(MBAinc == 1)
		{
			mpeg2video_macroblock_modes(slice, 
				video, 
				video->layerdata, 
				&mb_type, 
				&stwtype, 
				&stwclass,
        		&motion_type, 
				&mv_count, 
				&mv_format, 
				&dmv, 
				&mvscale, 
				&dct_type);

			if(slice->fault) return 1;

      		if(mb_type & MB_QUANT)
			{
        		qs = mpeg2slice_getbits(slice->slice_buffer, 5);

        		if(video->layerdata->mpeg2)
            	 	slice->quant_scale = video->layerdata->qscale_type ? mpeg2_non_linear_mquant_table[qs] : (qs << 1);
        		else 
					slice->quant_scale = qs;

        		if(video->layerdata->scalable_mode == SC_DP)
/* make sure quant_scale is valid */
          			slice->quant_scale = slice->quant_scale;
      		}

/* motion vectors */


/* decode forward motion vectors */
      		if((mb_type & MB_FORWARD) || ((mb_type & MB_INTRA) && video->conceal_mv))
			{
        		if(video->layerdata->mpeg2)
        			mpeg2video_motion_vectors(slice, 
						video, 
						PMV, 
						dmvector, 
						mv_field_sel, 
            			0, 
						mv_count, 
						mv_format, 
						video->h_forw_r_size, 
						video->v_forw_r_size, 
						dmv, 
						mvscale);
        		else
        		  	mpeg2video_motion_vector(slice, 
						video, 
						PMV[0][0], 
						dmvector, 
            			video->forw_r_size, 
						video->forw_r_size, 
						0, 
						0, 
						video->full_forw);
    		}
      		if(slice->fault) return 1;

/* decode backward motion vectors */
    		if(mb_type & MB_BACKWARD)
			{
        		if(video->layerdata->mpeg2)
        		  	mpeg2video_motion_vectors(slice, 
						video, 
						PMV, 
						dmvector, 
						mv_field_sel, 
            			1, 
						mv_count, 
						mv_format, 
						video->h_back_r_size, 
						video->v_back_r_size, 
						0, 
						mvscale);
        		else
        		  	mpeg2video_motion_vector(slice, 
						video, 
						PMV[0][1], 
						dmvector, 
            			video->back_r_size, 
						video->back_r_size, 
						0, 
						0, 
						video->full_back);
    		}

      		if(slice->fault) return 1;

/* remove marker_bit */
      		if((mb_type & MB_INTRA) && video->conceal_mv)
        		mpeg2slice_flushbits(slice->slice_buffer, 1);

/* macroblock_pattern */
      		if(mb_type & MB_PATTERN)
			{
        		cbp = mpeg2video_get_cbp(slice);
        		if(video->chroma_format == CHROMA422)
				{
/* coded_block_pattern_1 */
        		  	cbp = (cbp << 2) | mpeg2slice_getbits(slice->slice_buffer, 2); 
        		}
        		else
				if(video->chroma_format == CHROMA444)
				{
/* coded_block_pattern_2 */
        		  	cbp = (cbp << 6) | mpeg2slice_getbits(slice->slice_buffer, 6); 
        		}
    		}
    		else
        	  	cbp = (mb_type & MB_INTRA) ? ((1 << video->blk_cnt) - 1) : 0;

      		if(slice->fault) return 1;

/* decode blocks */
      		mpeg2video_clearblock(slice, 0, video->blk_cnt);
      		for(comp = 0; comp < video->blk_cnt; comp++)
			{
        		if(cbp & (1 << (video->blk_cnt - comp - 1)))
				{
          			if(mb_type & MB_INTRA)
					{
            			if(video->layerdata->mpeg2)
							mpeg2video_getmpg2intrablock(slice, video, video->layerdata, comp, dc_dct_pred);
            			else
							mpeg2video_getintrablock(slice, video->layerdata, comp, dc_dct_pred);
          			}
        			else 
					{
            		  	if(video->layerdata->mpeg2) 
					  		mpeg2video_getmpg2interblock(slice, video, video->layerdata, comp);
            		  	else           
					  		mpeg2video_getinterblock(slice, video->layerdata, comp);
        			}
        			if(slice->fault) return 1;
        		}
      		}

/* reset intra_dc predictors */
			if(!(mb_type & MB_INTRA))
        	  	dc_dct_pred[0] = dc_dct_pred[1] = dc_dct_pred[2] = 0;

/* reset motion vector predictors */
    		if((mb_type & MB_INTRA) && !video->conceal_mv)
			{
/* intra mb without concealment motion vectors */
        		PMV[0][0][0] = PMV[0][0][1] = PMV[1][0][0] = PMV[1][0][1] = 0;
        		PMV[0][1][0] = PMV[0][1][1] = PMV[1][1][0] = PMV[1][1][1] = 0;
    		}

    		if((video->pict_type == P_TYPE) && !(mb_type & (MB_FORWARD | MB_INTRA)))
			{
/* non-intra mb without forward mv in a P picture */
        		PMV[0][0][0] = PMV[0][0][1] = PMV[1][0][0] = PMV[1][0][1] = 0;

/* derive motion_type */
        		if(video->pict_struct == FRAME_PICTURE) 
					motion_type = MC_FRAME;
        		else
        		{
        			motion_type = MC_FIELD;
/* predict from field of same parity */
        			mv_field_sel[0][0] = (video->pict_struct == BOTTOM_FIELD);
        		}
      		}

    		if(stwclass == 4)
    		{
/* purely spatially predicted macroblock */
        		PMV[0][0][0] = PMV[0][0][1] = PMV[1][0][0] = PMV[1][0][1] = 0;
        		PMV[0][1][0] = PMV[0][1][1] = PMV[1][1][0] = PMV[1][1][1] = 0;
    		}
    	}
    	else 
		{
/* MBAinc!=1: skipped macroblock */
      		mpeg2video_clearblock(slice, 0, video->blk_cnt);

/* reset intra_dc predictors */
      		dc_dct_pred[0] = dc_dct_pred[1] = dc_dct_pred[2] = 0;

/* reset motion vector predictors */
      		if(video->pict_type == P_TYPE)
        		PMV[0][0][0] = PMV[0][0][1] = PMV[1][0][0] = PMV[1][0][1] = 0;

/* derive motion_type */
      		if(video->pict_struct == FRAME_PICTURE)
        		motion_type = MC_FRAME;
    		else
    		{
        		motion_type = MC_FIELD;
/* predict from field of same parity */
        		mv_field_sel[0][0] = mv_field_sel[0][1] = (video->pict_struct == BOTTOM_FIELD);
    		}

/* skipped I are spatial-only predicted, */
/* skipped P and B are temporal-only predicted */
      		stwtype = (video->pict_type == I_TYPE) ? 8 : 0;

/* clear MB_INTRA */
      		mb_type &= ~MB_INTRA;

/* no block data */
      		cbp = 0; 
    	}

    	SNRcbp = 0;

/* pixel coordinates of top left corner of current macroblock */
    	bx = 16 * (MBA % video->mb_width);
    	by = 16 * (MBA / video->mb_width);

/* motion compensation */
    	if(!(mb_type & MB_INTRA))
    	  	mpeg2video_reconstruct(video, 
				video->layerdata, 
				bx, 
				by, 
				mb_type, 
				motion_type, 
				PMV, 
				mv_field_sel, 
				dmvector, 
				stwtype);

/* copy or add block data into picture */
    	for(comp = 0; comp < video->blk_cnt; comp++)
		{
      		if((cbp | SNRcbp) & (1 << (video->blk_cnt - 1 - comp)))
			{
#ifdef HAVE_MMX
  	  			IDCT_mmx(slice->block[comp]);
#else
          		mpeg2video_idct_conversion(slice->block[comp]);
#endif
        		mpeg2video_addblock(slice, 
					video, 
					video->layerdata, 
					comp, 
					bx, 
					by, 
					dct_type, 
					(mb_type & MB_INTRA) == 0);
      		}
    	}

/* advance to next macroblock */
    	MBA++;
    	MBAinc--;
// Align the bytes?
//		mpeg2slice_showbits(slice->slice_buffer, 32);
//printf("debug slice 9 %d %x\n", MBAinc, mpeg2slice_showbits(slice->slice_buffer, 16));
  	}

	return 0;
}

void mpeg2_slice_loop(mpeg2_slice_t *slice)
{
	mpeg2video_t *video = slice->video;
	int result = 1;

	while(!slice->done)
	{
		pthread_mutex_lock(&(slice->input_lock));

		if(!slice->done)
		{
// Get a buffer to decode
			result = 1;
			pthread_mutex_lock(&(video->slice_lock));
			if(slice->buffer_step > 0)
			{
				while(slice->current_buffer <= slice->last_buffer)
				{
					if(!video->slice_buffers[slice->current_buffer].done &&
						slice->current_buffer <= slice->last_buffer)
					{
						result = 0;
						break;
					}
					slice->current_buffer += slice->buffer_step;
				}
			}
			else
			{
				while(slice->current_buffer >= slice->last_buffer)
				{
					if(!video->slice_buffers[slice->current_buffer].done &&
						slice->current_buffer >= slice->last_buffer)
					{
						result = 0;
						break;
					}
					slice->current_buffer += slice->buffer_step;
				}
			}

// Got one
			if(!result && slice->current_buffer >= 0 && slice->current_buffer < video->total_slice_buffers)
			{
				slice->slice_buffer = &(video->slice_buffers[slice->current_buffer]);
				slice->slice_buffer->done = 1;
				pthread_mutex_unlock(&(video->slice_lock));
				pthread_mutex_unlock(&(slice->input_lock));
				mpeg2_decode_slice(slice);
				pthread_mutex_unlock(&(slice->slice_buffer->completion_lock));
			}
			else
				pthread_mutex_unlock(&(video->slice_lock));
		}

// Empty mmx registers seperately for each CPU
#ifdef HAVE_MMX
		__asm__ __volatile__ ("emms");
#endif

		pthread_mutex_unlock(&(slice->output_lock));
	}
}

int mpeg2_new_slice_decoder(void *video, mpeg2_slice_t *slice)
{
	pthread_attr_t  attr;
	struct sched_param param;
	pthread_mutexattr_t mutex_attr;

	slice->video = video;
	slice->done = 0;
	pthread_mutexattr_init(&mutex_attr);
	pthread_mutex_init(&(slice->input_lock), &mutex_attr);
	pthread_mutex_lock(&(slice->input_lock));
	pthread_mutex_init(&(slice->output_lock), &mutex_attr);
	pthread_mutex_lock(&(slice->output_lock));

	pthread_attr_init(&attr);
	pthread_create(&(slice->tid), &attr, (void*)mpeg2_slice_loop, slice);

	return 0;
}

int mpeg2_delete_slice_decoder(mpeg2_slice_t *slice)
{
	slice->done = 1;
	pthread_mutex_unlock(&(slice->input_lock));
	pthread_join(slice->tid, 0);
	pthread_mutex_destroy(&(slice->input_lock));
	pthread_mutex_destroy(&(slice->output_lock));
	return 0;
}
